#include "block.h"
class _BlockStack ;
/**********************************************************/
/* This code is for PLDI-15 Artifact Evaluation only      */
/* and will be released with further copyright information*/
/* File: Basic sequential recursive version of knapsack  */
/**********************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <algorithm>
#include <math.h>
#include <stdarg.h>
#include <string.h>
/********************Contents in getoptions************************************************/
/*
 * Function to evaluate argv[]. specs is a 0 terminated array of command
 * line options and types an array that stores the type as one of
 */
#define INTARG 1
#define DOUBLEARG 2
#define LONGARG 3
#define BOOLARG 4
#define STRINGARG 5
#define BENCHMARK 6
/*
 * for each specifier. Benchmark is specific for cilk samples.
 * -benchmark or -benchmark medium sets integer to 2
 * -benchmark short returns 1
 * -benchmark long returns 2
 * a boolarg is set to 1 if the specifier appears in the option list ow. 0
 * The variables must be given in the same order as specified in specs.
 */
void get_options(int argc,char *argv[],char *specs[],int *types,... );
/* Used by example programs to evaluate command line options */

void get_options(int argc,char *argv[],char *specs[],int *types,... )
{
  va_list ap;
  int type;
  int i;
  int *intval;
  double *doubleval;
  long *longval;
  char *stringval;
  va_start(ap,types);
  while(((type =  *(types++)) != 0) && (specs != 0)){
    switch(type){
      case 1:
{
        intval = (va_arg(ap,int *));
        for (i = 1; i < (argc - 1); i++) 
//for (i = 0; i < (argc - 1); i++)
          if (!(strcmp(argv[i],specs[0]))) {
             *intval = atoi(argv[i + 1]);
            argv[i][0] = 0;
            argv[i + 1][0] = 0;
          }
        break; 
      }
      case 2:
{
        doubleval = (va_arg(ap,double *));
        for (i = 1; i < (argc - 1); i++) 
//for (i = 0; i < (argc - 1); i++)
          if (!(strcmp(argv[i],specs[0]))) {
             *doubleval = atof(argv[i + 1]);
            argv[i][0] = 0;
            argv[i + 1][0] = 0;
          }
        break; 
      }
      case 3:
{
        longval = (va_arg(ap,long *));
        for (i = 1; i < (argc - 1); i++) 
//for (i = 0; i < (argc - 1); i++)
          if (!(strcmp(argv[i],specs[0]))) {
             *longval = atol(argv[i + 1]);
            argv[i][0] = 0;
            argv[i + 1][0] = 0;
          }
        break; 
      }
      case 4:
{
        intval = (va_arg(ap,int *));
         *intval = 0;
        for (i = 1; i < argc; i++) 
//for (i = 0; i < argc; i++)
          if (!(strcmp(argv[i],specs[0]))) {
             *intval = 1;
            argv[i][0] = 0;
          }
        break; 
      }
      case 5:
{
        stringval = (va_arg(ap,char *));
        for (i = 1; i < (argc - 1); i++) 
//for (i = 0; i < (argc - 1); i++)
          if (!(strcmp(argv[i],specs[0]))) {
            strcpy(stringval,argv[i + 1]);
            argv[i][0] = 0;
            argv[i + 1][0] = 0;
          }
        break; 
      }
      case 6:
{
        intval = (va_arg(ap,int *));
         *intval = 0;
        for (i = 1; i < argc; i++) {
//for (i = 0; i < argc; i++){
          if (!(strcmp(argv[i],specs[0]))) {
             *intval = 2;
            if ((i + 1) < argc) {
              if (!(strcmp(argv[i + 1],"short"))) 
                 *intval = 1;
              if (!(strcmp(argv[i + 1],"medium"))) 
                 *intval = 2;
              if (!(strcmp(argv[i + 1],"long"))) 
                 *intval = 3;
              argv[i + 1][0] = 0;
            }
            argv[i][0] = 0;
          }
        }
        break; 
      }
    }
    specs++;
  }
  va_end(ap);
  for (i = 1; i < argc; i++) 
//for (i = 0; i < argc; i++)
    if (argv[i][0] != 0) 
      printf("\nInvalid option: %s\n",argv[i]);
}
/********************************************************************/
using namespace std;
int _expandSize = 16;
int _minBlockSize = 16;
/* every item in the knapsack has a weight and a value */
#define MAX_ITEMS 256

struct item 
{
  int value;
  int weight;
}
;
int best_so_far = -2147483647 - 1;

int compare(struct item *a,struct item *b)
{
  double c = ((((double )(a -> item::value)) / (a -> item::weight)) - (((double )(b -> item::value)) / (b -> item::weight)));
  if (c > 0) 
    return -1;
  if (c < 0) 
    return 1;
  return 0;
}

int read_input(const char *filename,struct item *items,int *capacity,int *n)
{
  int i;
  FILE *f;
  if (filename == 0L) 
    filename = "\000";
  f = fopen(filename,"r");
  if (f == 0L) {
    fprintf(stderr,"open_input(\"%s\") failed\n",filename);
    return -1;
  }
/* format of the input: #items capacity value1 weight1 ... */
  fscanf(f,"%d",n);
  fscanf(f,"%d",capacity);
  for (i = 0; i <  *n; ++i) 
    fscanf(f,"%d %d",&items[i].item::value,&items[i].item::weight);
  fclose(f);
/* sort the items on decreasing order of value/weight */
/* cilk2c is fascist in dealing with pointers, whence the ugly cast */
  qsort(items,( *n),(sizeof(struct item )),((int (*)(const void *, const void *))compare));
  return 0;
}
void knapsack_block_bf(struct item *e,int n,int *max_ret,_BlockStack *_stack,int _depth);
void knapsack_block_df0(struct item *e,int n,int *max_ret,_BlockStack *_stack,int _depth);
void knapsack_block_df1(struct item *e,int n,int *max_ret,_BlockStack *_stack,int _depth);
void knapsack(struct item *e,int n,int c0,int v0,int c1,int v1,int *max_ret);
/*
 * return the optimal solution for n items (first is e) and
 * capacity c. Value so far is v.
 */

void knapsack(struct item *e,int n,int c0,int v0,int c1,int v1,int *max_ret)
{
  if (c0 < 0) {
     *max_ret = std::max< int  > ( *max_ret,(-2147483647 - 1));
  }
  else if ((n == 0) || (c0 == 0)) {
     *max_ret = std::max< int  > ( *max_ret,v0);
  }
  else {
    knapsack((e + 1),(n - 1),c0,v0,(c0 - (e -> item::weight)),(v0 + (e -> item::value)),max_ret);
  }
  if (c1 < 0) {
     *max_ret = std::max< int  > ( *max_ret,(-2147483647 - 1));
  }
  else if ((n == 0) || (c1 == 0)) {
     *max_ret = std::max< int  > ( *max_ret,v1);
  }
  else {
    knapsack((e + 1),(n - 1),c1,v1,(c1 - (e -> item::weight)),(v1 + (e -> item::value)),max_ret);
  }
}

int usage()
{
  fprintf(stderr,"\nUsage: knapsack [<cilk-options>] [-f filename] [-benchmark] [-h]\n\n");
  fprintf(stderr,"The 0-1-Knapsack is a standard combinatorial optimization problem: ``A\n");
  fprintf(stderr,"thief robbing a store finds n items; the ith item is worth v_i dollars\n");
  fprintf(stderr,"and weighs w_i pounds, where v_i and w_i are integers. He wants to take\n");
  fprintf(stderr,"as valuable a load as possible, but he can carry at most W pounds in\n");
  fprintf(stderr,"his knapsack for some integer W. What items should he take\?\'\'\n\n");
  return -1;
}
char *specifiers[] = {("-f"), ("-benchmark"), ("-h"), ("-b"), (0)};
int opt_types[] = {(5), (6), (4), (1), (0)};

int main(int argc,char *argv[])
{
/* array of items */
  struct item items[256UL];
  int n;
  int capacity;
  int sol;
  int benchmark;
  int help;
  int buf;
  char filename[100UL];
  buf = 0;
  sol = 0;
/* standard benchmark options */
  strcpy(filename,"knapsack-example2.input");
  get_options(argc,argv,specifiers,opt_types,filename,&benchmark,&help,&buf);
//    printf("This is buffer size:%d\n", buf);
  if (buf) {
    _expandSize = buf;
    printf("_expandSize: %d\n",_expandSize);
  }
  _minBlockSize = 2;
  if (help) 
    return usage();
  if (benchmark) {
    switch(benchmark){
/* short benchmark options -- a little work */
      case 1:
{
        strcpy(filename,"knapsack-example1.input");
        break; 
      }
/* standard benchmark options */
      case 2:
{
        strcpy(filename,"knapsack-example2.input");
        break; 
      }
/* long benchmark options -- a lot of work */
      case 3:
{
        strcpy(filename,"knapsack-example3.input");
        break; 
      }
    }
  }
  if ((read_input(filename,items,&capacity,&n))) 
    return 1;
/* base case: full knapsack or no items */
  if (capacity < 0) {
    sol = std::max< int  > (sol,(-2147483647 - 1));
    return 0;
  }
  if ((n == 0) || (capacity == 0)) {
    sol = std::max< int  > (sol,(0));
/* feasible solution, with value v */
    return 0;
  }
  int c0 = capacity;
  int v0 = 0;
  int c1 = (capacity - items[0].item::weight);
  int v1 = items[0].item::value;
  
#pragma recursionCallEntry
  knapsack((items + 1),(n - 1),c0,v0,c1,v1,&sol);
  printf("\nExample: knapsack\n");
  printf("options: problem-file = %s\n\n",filename);
  printf("Best value is %d\n\n",sol);
  return 0;
}
