#include "block.h"
class _BlockStack ;
/**********************************************************************************************/
/*  This program is part of the Barcelona OpenMP Tasks Suite                                  */
/*  Copyright (C) 2009 Barcelona Supercomputing Center - Centro Nacional de Supercomputacion  */
/*  Copyright (C) 2009 Universitat Politecnica de Catalunya                                   */
/**********************************************************************************************/
/*
 * Copyright (c) 2007 The Unbalanced Tree Search (UTS) Project Team:
 * -----------------------------------------------------------------
 *
 *  This file is part of the unbalanced tree search benchmark.  This
 *  project is licensed under the MIT Open Source license.  See the LICENSE
 *  file for copyright and licensing information.
 *
 *  UTS is a collaborative project between researchers at the University of
 *  Maryland, the University of North Carolina at Chapel Hill, and the Ohio
 *  State University.
 *
 * University of Maryland:
 *   Chau-Wen Tseng(1)  <tseng at cs.umd.edu>
 *
 * University of North Carolina, Chapel Hill:
 *   Jun Huan         <huan,
 *   Jinze Liu         liu,
 *   Stephen Olivier   olivier,
 *   Jan Prins*        prins at cs.umd.edu>
 *
 * The Ohio State University:
 *   James Dinan      <dinan,
 *   Gerald Sabin      sabin,
 *   P. Sadayappan*    saday at cse.ohio-state.edu>
 *
 * Supercomputing Research Center
 *   D. Pryor
 *
 * (1) - indicates project PI
 *
 * UTS Recursive Depth-First Search (DFS) version developed by James Dinan
 *
 * Adapted for OpenMP 3.0 Task-based version by Stephen Olivier
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <omp.h>
#include <sys/time.h>
////////////////////////////////#include "head.h"
#define bots_message(msg, ...) \
   {\
        fprintf(stdout, msg , ##__VA_ARGS__);\
   }
#define bots_debug(msg, ...) \
   {\
       fprintf(stdout, msg , ##__VA_ARGS__);\
   }
////////////////////////////////
//////////////////////////////#include "brg_types.h"
#ifndef BRG_TYPES_H
#define BRG_TYPES_H
#include <sys/types.h>
#if defined(__cplusplus)
extern "C" {
#endif
#include <limits.h>
/* Try this if you you get an error from one of the typedefs below */
#ifdef BRG_STD_TYPES
#  define BRG_UI8
#  define BRG_UI16
#  define BRG_UI32
#    define li_32(h) 0x##h##u
#  define BRG_UI64
#    define li_64(h) 0x##h##u
#endif /* BRG_C99_TYPES */
#ifndef BRG_UI8
#  define BRG_UI8
#  if UCHAR_MAX == 255u
typedef unsigned char uint_8t;
#  else
#    error Please define uint_8t as an 8-bit unsigned integer type in brg_types.h
#  endif
#endif
#ifndef BRG_UI16
#  define BRG_UI16
#  if USHRT_MAX == 65535u
typedef unsigned short uint_16t;
#  else
#    error Please define uint_16t as a 16-bit unsigned short type in brg_types.h
#  endif
#endif
#ifndef BRG_UI32
#  define BRG_UI32
#  if UINT_MAX == 4294967295u
#    define li_32(h) 0x##h##u
typedef unsigned int uint_32t;
#  elif ULONG_MAX == 4294967295u
#    define li_32(h) 0x##h##ul
#  elif defined( _CRAY )
#    error This code needs 32-bit data types, which Cray machines do not provide
#  else
#    error Please define uint_32t as a 32-bit unsigned integer type in brg_types.h
#  endif
#endif
#ifndef BRG_UI64
#  if defined( __BORLANDC__ ) && !defined( __MSDOS__ )
#    define BRG_UI64
#    define li_64(h) 0x##h##ull
#  elif defined( _MSC_VER ) && ( _MSC_VER < 1300 )    /* 1300 == VC++ 7.0 */
#    define BRG_UI64
#    define li_64(h) 0x##h##ui64
#  elif defined( __sun ) && defined(ULONG_MAX) && ULONG_MAX == 0xfffffffful
#    define BRG_UI64
#    define li_64(h) 0x##h##ull
#  elif defined( UINT_MAX ) && UINT_MAX > 4294967295u
#    if UINT_MAX == 18446744073709551615u
#      define BRG_UI64
#      define li_64(h) 0x##h##u
#    endif
#  elif defined( ULONG_MAX ) && ULONG_MAX > 4294967295u
#    if ULONG_MAX == 18446744073709551615ul
#      define BRG_UI64
#      define li_64(h) 0x##h##ul
typedef unsigned long uint_64t;
#    endif
#  elif defined( ULLONG_MAX ) && ULLONG_MAX > 4294967295u
#    if ULLONG_MAX == 18446744073709551615ull
#      define BRG_UI64
#      define li_64(h) 0x##h##ull
#    endif
#  elif defined( ULONG_LONG_MAX ) && ULONG_LONG_MAX > 4294967295u
#    if ULONG_LONG_MAX == 18446744073709551615ull
#      define BRG_UI64
#      define li_64(h) 0x##h##ull
#    endif
#  endif
#endif
#if defined( NEED_UINT_64T ) && !defined( BRG_UI64 )
#  error Please define uint_64t as an unsigned 64 bit type in brg_types.h
#endif
#ifndef RETURN_VALUES
#  define RETURN_VALUES
#  if defined( DLL_EXPORT )
#    if defined( _MSC_VER ) || defined ( __INTEL_COMPILER )
#      define VOID_RETURN    __declspec( dllexport ) void __stdcall
#      define INT_RETURN     __declspec( dllexport ) int  __stdcall
#    elif defined( __GNUC__ )
#      define VOID_RETURN    __declspec( __dllexport__ ) void
#      define INT_RETURN     __declspec( __dllexport__ ) int
#    else
#      error Use of the DLL is only available on the Microsoft, Intel and GCC compilers
#    endif
#  elif defined( DLL_IMPORT )
#    if defined( _MSC_VER ) || defined ( __INTEL_COMPILER )
#      define VOID_RETURN    __declspec( dllimport ) void __stdcall
#      define INT_RETURN     __declspec( dllimport ) int  __stdcall
#    elif defined( __GNUC__ )
#      define VOID_RETURN    __declspec( __dllimport__ ) void
#      define INT_RETURN     __declspec( __dllimport__ ) int
#    else
#      error Use of the DLL is only available on the Microsoft, Intel and GCC compilers
#    endif
#  elif defined( __WATCOMC__ )
#    define VOID_RETURN  void __cdecl
#    define INT_RETURN   int  __cdecl
#  else
#    define VOID_RETURN  void
#    define INT_RETURN   int
#  endif
#endif
/*  These defines are used to declare buffers in a way that allows
    faster operations on longer variables to be used.  In all these
    defines 'size' must be a power of 2 and >= 8
    dec_unit_type(size,x)       declares a variable 'x' of length
                                'size' bits
    dec_bufr_type(size,bsize,x) declares a buffer 'x' of length 'bsize'
                                bytes defined as an array of variables
                                each of 'size' bits (bsize must be a
                                multiple of size / 8)
    ptr_cast(x,size)            casts a pointer to a pointer to a
                                varaiable of length 'size' bits
*/
#define ui_type(size)               uint_##size##t
#define dec_unit_type(size,x)       typedef ui_type(size) x
#define dec_bufr_type(size,bsize,x) typedef ui_type(size) x[bsize / (size >> 3)]
#define ptr_cast(x,size)            ((ui_type(size)*)(x))
#if defined(__cplusplus)
}
#endif
#endif
//////////////////////////////////////////////////////////////
//////////////////////////////////// #include "brg_sha1.h"
#ifndef _SHA1_H
#define _SHA1_H
#include <stdlib.h>
#define SHA1_BLOCK_SIZE  64
#define SHA1_DIGEST_SIZE 20
#if defined(__cplusplus)
extern "C"
{
#endif
/** BEGIN: UTS RNG Harness **/
#define POS_MASK    0x7fffffff
#define HIGH_BITS   0x80000000
#define sha1_context sha1_ctx_s
typedef u_int8_t RNG_state;
typedef u_int32_t uint32;
/**********************************/
/* random number generator state  */
/**********************************/

struct state_t 
{
  u_int8_t state[20UL];
}
;
/***************************************/
/* random number generator operations  */
/***************************************/
extern "C" { void rng_init(RNG_state *state,int seed); }
extern "C" { void rng_spawn(RNG_state *mystate,RNG_state *newstate,int spawnNumber); }
extern "C" { int rng_rand(RNG_state *mystate); }
extern "C" { double rng_toProb(int n); }
extern "C" { int rng_nextrand(RNG_state *mystate); }
extern "C" { char *rng_showstate(RNG_state *state,char *s); }
extern "C" { void rng_showtype(); }
/** END: UTS RNG Harness **/
/* type to hold the SHA256 context  */

struct sha1_ctx_s 
{
  uint_32t count[2UL];
  uint_32t hash[5UL];
  uint_32t wbuf[16UL];
}
;
typedef struct sha1_ctx_s sha1_ctx;
/* Note that these prototypes are the same for both bit and */
/* byte oriented implementations. However the length fields */
/* are in bytes or bits as appropriate for the version used */
/* and bit sequences are input as arrays of bytes in which  */
/* bit sequences run from the most to the least significant */
/* end of each byte                                         */
extern "C" { void sha1_compile(sha1_ctx ctx[1UL]); }
extern "C" { void sha1_begin(sha1_ctx ctx[1UL]); }
extern "C" { void sha1_hash(const unsigned char data[],unsigned long len,sha1_ctx ctx[1UL]); }
extern "C" { void sha1_end(unsigned char hval[],sha1_ctx ctx[1UL]); }
extern "C" { void sha1(unsigned char hval[],const unsigned char data[],unsigned long len); }
#if defined(__cplusplus)
}
#endif
#endif
////////////////////////////////////////////////////////////////
///////////////////////////////////////////#include "brg_endian.h"
#ifndef BRG_ENDIAN_H
#define BRG_ENDIAN_H
#define IS_BIG_ENDIAN      4321 /* byte 0 is most significant (mc68k) */
#define IS_LITTLE_ENDIAN   1234 /* byte 0 is least significant (i386) */
/* Include files where endian defines and byteswap functions may reside */
#if defined( __FreeBSD__ ) || defined( __OpenBSD__ ) || defined( __NetBSD__ )
#  include <sys/endian.h>
#elif defined( BSD ) && ( BSD >= 199103 ) || defined( __APPLE__ ) || \
      defined( __CYGWIN32__ ) || defined( __DJGPP__ ) || defined( __osf__ )
#  include <machine/endian.h>
#elif defined( __linux__ ) || defined( __GNUC__ ) || defined( __GNU_LIBRARY__ )
#  if !defined( __MINGW32__ )
#    include <endian.h>
#    if !defined( __BEOS__ )
#      include <byteswap.h>
#    endif
#  endif
#endif
/* Now attempt to set the define for platform byte order using any  */
/* of the four forms SYMBOL, _SYMBOL, __SYMBOL & __SYMBOL__, which  */
/* seem to encompass most endian symbol definitions                 */
#if defined( BIG_ENDIAN ) && defined( LITTLE_ENDIAN )
#  if defined( BYTE_ORDER ) && BYTE_ORDER == BIG_ENDIAN
#    define PLATFORM_BYTE_ORDER IS_BIG_ENDIAN
#  elif defined( BYTE_ORDER ) && BYTE_ORDER == LITTLE_ENDIAN
#    define PLATFORM_BYTE_ORDER IS_LITTLE_ENDIAN
#  endif
#elif defined( BIG_ENDIAN )
#  define PLATFORM_BYTE_ORDER IS_BIG_ENDIAN
#elif defined( LITTLE_ENDIAN )
#  define PLATFORM_BYTE_ORDER IS_LITTLE_ENDIAN
#endif
#if defined( _BIG_ENDIAN ) && defined( _LITTLE_ENDIAN )
#  if defined( _BYTE_ORDER ) && _BYTE_ORDER == _BIG_ENDIAN
#    define PLATFORM_BYTE_ORDER IS_BIG_ENDIAN
#  elif defined( _BYTE_ORDER ) && _BYTE_ORDER == _LITTLE_ENDIAN
#    define PLATFORM_BYTE_ORDER IS_LITTLE_ENDIAN
#  endif
#elif defined( _BIG_ENDIAN )
#  define PLATFORM_BYTE_ORDER IS_BIG_ENDIAN
#elif defined( _LITTLE_ENDIAN )
#  define PLATFORM_BYTE_ORDER IS_LITTLE_ENDIAN
#endif
#if defined( __BIG_ENDIAN ) && defined( __LITTLE_ENDIAN )
#  if defined( __BYTE_ORDER ) && __BYTE_ORDER == __BIG_ENDIAN
#    define PLATFORM_BYTE_ORDER IS_BIG_ENDIAN
#  elif defined( __BYTE_ORDER ) && __BYTE_ORDER == __LITTLE_ENDIAN
#    define PLATFORM_BYTE_ORDER IS_LITTLE_ENDIAN
#  endif
#elif defined( __BIG_ENDIAN )
#  define PLATFORM_BYTE_ORDER IS_BIG_ENDIAN
#elif defined( __LITTLE_ENDIAN )
#  define PLATFORM_BYTE_ORDER IS_LITTLE_ENDIAN
#endif
#if defined( __BIG_ENDIAN__ ) && defined( __LITTLE_ENDIAN__ )
#  if defined( __BYTE_ORDER__ ) && __BYTE_ORDER__ == __BIG_ENDIAN__
#    define PLATFORM_BYTE_ORDER IS_BIG_ENDIAN
#  elif defined( __BYTE_ORDER__ ) && __BYTE_ORDER__ == __LITTLE_ENDIAN__
#    define PLATFORM_BYTE_ORDER IS_LITTLE_ENDIAN
#  endif
#elif defined( __BIG_ENDIAN__ )
#  define PLATFORM_BYTE_ORDER IS_BIG_ENDIAN
#elif defined( __LITTLE_ENDIAN__ )
#  define PLATFORM_BYTE_ORDER IS_LITTLE_ENDIAN
#endif
/*  if the platform byte order could not be determined, then try to */
/*  set this define using common machine defines                    */
#if !defined(PLATFORM_BYTE_ORDER)
#if   defined( __alpha__ ) || defined( __alpha ) || defined( i386 )       || \
      defined( __i386__ )  || defined( _M_I86 )  || defined( _M_IX86 )    || \
      defined( __OS2__ )   || defined( sun386 )  || defined( __TURBOC__ ) || \
      defined( vax )       || defined( vms )     || defined( VMS )        || \
      defined( __VMS )     || defined( _M_X64 )
#  define PLATFORM_BYTE_ORDER IS_LITTLE_ENDIAN
#elif defined( AMIGA )   || defined( applec )    || defined( __AS400__ )  || \
      defined( _CRAY )   || defined( __hppa )    || defined( __hp9000 )   || \
      defined( ibm370 )  || defined( mc68000 )   || defined( m68k )       || \
      defined( __MRC__ ) || defined( __MVS__ )   || defined( __MWERKS__ ) || \
      defined( sparc )   || defined( __sparc)    || defined( SYMANTEC_C ) || \
      defined( __VOS__ ) || defined( __TIGCC__ ) || defined( __TANDEM )   || \
      defined( THINK_C ) || defined( __VMCMS__ )
#  define PLATFORM_BYTE_ORDER IS_BIG_ENDIAN
#elif 0     /* **** EDIT HERE IF NECESSARY **** */
#  define PLATFORM_BYTE_ORDER IS_LITTLE_ENDIAN
#elif 0     /* **** EDIT HERE IF NECESSARY **** */
#  define PLATFORM_BYTE_ORDER IS_BIG_ENDIAN
#else
#  error Please edit lines 126 or 128 in brg_endian.h to set the platform byte order
#endif
#endif
#endif
////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////brg_sha1.cpp
#if defined(__cplusplus)
extern "C"
{
#endif
/** BEGIN: UTS RNG Harness **/

extern "C" { void rng_init(RNG_state *newstate,int seed)
{
  struct sha1_ctx_s ctx;
  struct state_t gen;
  int i;
  for (i = 0; i < 16; i++) 
    gen.state_t::state[i] = 0;
  gen.state_t::state[16] = ((u_int8_t )(0xFF & (seed >> 24)));
  gen.state_t::state[17] = ((u_int8_t )(0xFF & (seed >> 16)));
  gen.state_t::state[18] = ((u_int8_t )(0xFF & (seed >> 8)));
  gen.state_t::state[19] = ((u_int8_t )(0xFF & (seed >> 0)));
  sha1_begin(&ctx);
  sha1_hash(gen.state_t::state,20,&ctx);
  sha1_end(newstate,&ctx);
}
 }

extern "C" { void rng_spawn(RNG_state *mystate,RNG_state *newstate,int spawnnumber)
{
  struct sha1_ctx_s ctx;
  u_int8_t bytes[4UL];
  bytes[0] = ((u_int8_t )(0xFF & (spawnnumber >> 24)));
  bytes[1] = ((u_int8_t )(0xFF & (spawnnumber >> 16)));
  bytes[2] = ((u_int8_t )(0xFF & (spawnnumber >> 8)));
  bytes[3] = ((u_int8_t )(0xFF & spawnnumber));
  sha1_begin(&ctx);
  sha1_hash(mystate,20,&ctx);
  sha1_hash(bytes,4,&ctx);
  sha1_end(newstate,&ctx);
}
 }

extern "C" { int rng_rand(RNG_state *mystate)
{
  int r;
  uint32 b = ((((mystate[16] << 24) | (mystate[17] << 16)) | (mystate[18] << 8)) | (mystate[19] << 0));
  b = (b & 0x7fffffff);
  r = ((int )b);
//bots_debug("b: %d\t, r: %d\n", b, r);
  return r;
}
 }
// Interpret 32 bit positive integer as value on [0,1)

extern "C" { double rng_toProb(int n)
{
  if (n < 0) {
    printf("*** toProb: rand n = %d out of range\n",n);
  }
  return (n < 0)?0.0 : (((double )n) / 2147483648.0);
}
 }

extern "C" { int rng_nextrand(RNG_state *mystate)
{
  struct sha1_ctx_s ctx;
  int r;
  uint32 b;
  sha1_begin(&ctx);
  sha1_hash(mystate,20,&ctx);
  sha1_end(mystate,&ctx);
  b = ((((mystate[16] << 24) | (mystate[17] << 16)) | (mystate[18] << 8)) | (mystate[19] << 0));
  b = (b & 0x7fffffff);
  r = ((int )b);
  return r;
}
 }
/* condense state into string to display during debugging */

extern "C" { char *rng_showstate(RNG_state *state,char *s)
{
  sprintf(s,"%.2X%.2X...",state[0],state[1]);
  return s;
}
 }
/* describe random number generator type into string */

extern "C" { void rng_showtype()
{
{
    fprintf(stdout,"SHA-1 (state size = %luB)\n",sizeof(struct state_t ));
  };
}
 }
/** END: UTS RNG Harness **/
#if defined( _MSC_VER ) && ( _MSC_VER > 800 )
#endif
#if 0 && defined(_MSC_VER)
#define rotl32  _lrotl
#define rotr32  _lrotr
#else
#define rotl32(x,n)   (((x) << n) | ((x) >> (32 - n)))
#define rotr32(x,n)   (((x) >> n) | ((x) << (32 - n)))
#endif
#if !defined(bswap_32)
#define bswap_32(x) ((rotr32((x), 24) & 0x00ff00ff) | (rotr32((x), 8) & 0xff00ff00))
#endif
#if (PLATFORM_BYTE_ORDER == IS_LITTLE_ENDIAN)
#define SWAP_BYTES
#else
#undef  SWAP_BYTES
#endif
#if defined(SWAP_BYTES)
#define bsw_32(p,n) \
    { int _i = (n); while(_i--) ((uint_32t*)p)[_i] = bswap_32(((uint_32t*)p)[_i]); }
#else
#define bsw_32(p,n)
#endif
#define SHA1_MASK   (SHA1_BLOCK_SIZE - 1)
#if 0
#define ch(x,y,z)       (((x) & (y)) ^ (~(x) & (z)))
#define parity(x,y,z)   ((x) ^ (y) ^ (z))
#define maj(x,y,z)      (((x) & (y)) ^ ((x) & (z)) ^ ((y) & (z)))
#else   /* Discovered by Rich Schroeppel and Colin Plumb   */
#define ch(x,y,z)       ((z) ^ ((x) & ((y) ^ (z))))
#define parity(x,y,z)   ((x) ^ (y) ^ (z))
#define maj(x,y,z)      (((x) & (y)) | ((z) & ((x) ^ (y))))
#endif
/* Compile 64 bytes of hash data into SHA1 context. Note    */
/* that this routine assumes that the byte order in the     */
/* ctx->wbuf[] at this point is in such an order that low   */
/* address bytes in the ORIGINAL byte stream will go in     */
/* this buffer to the high end of 32-bit words on BOTH big  */
/* and little endian systems                                */
#ifdef ARRAY
#define q(v,n)  v[n]
#else
#define q(v,n)  v##n
#endif
#define one_cycle(v,a,b,c,d,e,f,k,h)            \
    q(v,e) += rotr32(q(v,a),27) +               \
              f(q(v,b),q(v,c),q(v,d)) + k + h;  \
    q(v,b)  = rotr32(q(v,b), 2)
#define five_cycle(v,f,k,i)                 \
    one_cycle(v, 0,1,2,3,4, f,k,hf(i  ));   \
    one_cycle(v, 4,0,1,2,3, f,k,hf(i+1));   \
    one_cycle(v, 3,4,0,1,2, f,k,hf(i+2));   \
    one_cycle(v, 2,3,4,0,1, f,k,hf(i+3));   \
    one_cycle(v, 1,2,3,4,0, f,k,hf(i+4))

extern "C" { void sha1_compile(sha1_ctx ctx[1UL])
{
  uint_32t *w = (ctx -> sha1_ctx_s::wbuf);
#ifdef ARRAY
#else
  uint_32t v0;
  uint_32t v1;
  uint_32t v2;
  uint_32t v3;
  uint_32t v4;
  v0 = (ctx -> sha1_ctx_s::hash)[0];
  v1 = (ctx -> sha1_ctx_s::hash)[1];
  v2 = (ctx -> sha1_ctx_s::hash)[2];
  v3 = (ctx -> sha1_ctx_s::hash)[3];
  v4 = (ctx -> sha1_ctx_s::hash)[4];
#endif
#define hf(i)   w[i]
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + (v3 ^ (v1 & (v2 ^ v3)))) + 0x5a827999) + w[0]);
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + (v2 ^ (v0 & (v1 ^ v2)))) + 0x5a827999) + w[0 + 1]);
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + (v1 ^ (v4 & (v0 ^ v1)))) + 0x5a827999) + w[0 + 2]);
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + (v0 ^ (v3 & (v4 ^ v0)))) + 0x5a827999) + w[0 + 3]);
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + (v4 ^ (v2 & (v3 ^ v4)))) + 0x5a827999) + w[0 + 4]);
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + (v3 ^ (v1 & (v2 ^ v3)))) + 0x5a827999) + w[5]);
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + (v2 ^ (v0 & (v1 ^ v2)))) + 0x5a827999) + w[5 + 1]);
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + (v1 ^ (v4 & (v0 ^ v1)))) + 0x5a827999) + w[5 + 2]);
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + (v0 ^ (v3 & (v4 ^ v0)))) + 0x5a827999) + w[5 + 3]);
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + (v4 ^ (v2 & (v3 ^ v4)))) + 0x5a827999) + w[5 + 4]);
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + (v3 ^ (v1 & (v2 ^ v3)))) + 0x5a827999) + w[10]);
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + (v2 ^ (v0 & (v1 ^ v2)))) + 0x5a827999) + w[10 + 1]);
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + (v1 ^ (v4 & (v0 ^ v1)))) + 0x5a827999) + w[10 + 2]);
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + (v0 ^ (v3 & (v4 ^ v0)))) + 0x5a827999) + w[10 + 3]);
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + (v4 ^ (v2 & (v3 ^ v4)))) + 0x5a827999) + w[10 + 4]);
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + (v3 ^ (v1 & (v2 ^ v3)))) + 0x5a827999) + w[15]);
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
#undef  hf
#define hf(i) (w[(i) & 15] = rotl32(                    \
                 w[((i) + 13) & 15] ^ w[((i) + 8) & 15] \
               ^ w[((i) +  2) & 15] ^ w[(i) & 15], 1))
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + (v2 ^ (v0 & (v1 ^ v2)))) + 0x5a827999) + (w[16 & 15] = (((((w[16 + 13 & 15] ^ w[16 + 8 & 15]) ^ w[16 + 2 & 15]) ^ w[16 & 15]) << 1) | ((((w[16 + 13 & 15] ^ w[16 + 8 & 15]) ^ w[16 + 2 & 15]) ^ w[16 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + (v1 ^ (v4 & (v0 ^ v1)))) + 0x5a827999) + (w[17 & 15] = (((((w[17 + 13 & 15] ^ w[17 + 8 & 15]) ^ w[17 + 2 & 15]) ^ w[17 & 15]) << 1) | ((((w[17 + 13 & 15] ^ w[17 + 8 & 15]) ^ w[17 + 2 & 15]) ^ w[17 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + (v0 ^ (v3 & (v4 ^ v0)))) + 0x5a827999) + (w[18 & 15] = (((((w[18 + 13 & 15] ^ w[18 + 8 & 15]) ^ w[18 + 2 & 15]) ^ w[18 & 15]) << 1) | ((((w[18 + 13 & 15] ^ w[18 + 8 & 15]) ^ w[18 + 2 & 15]) ^ w[18 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + (v4 ^ (v2 & (v3 ^ v4)))) + 0x5a827999) + (w[19 & 15] = (((((w[19 + 13 & 15] ^ w[19 + 8 & 15]) ^ w[19 + 2 & 15]) ^ w[19 & 15]) << 1) | ((((w[19 + 13 & 15] ^ w[19 + 8 & 15]) ^ w[19 + 2 & 15]) ^ w[19 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 ^ v2) ^ v3)) + 0x6ed9eba1) + (w[20 & 15] = (((((w[20 + 13 & 15] ^ w[20 + 8 & 15]) ^ w[20 + 2 & 15]) ^ w[20 & 15]) << 1) | ((((w[20 + 13 & 15] ^ w[20 + 8 & 15]) ^ w[20 + 2 & 15]) ^ w[20 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 ^ v1) ^ v2)) + 0x6ed9eba1) + (w[20 + 1 & 15] = (((((w[20 + 1 + 13 & 15] ^ w[20 + 1 + 8 & 15]) ^ w[20 + 1 + 2 & 15]) ^ w[20 + 1 & 15]) << 1) | ((((w[20 + 1 + 13 & 15] ^ w[20 + 1 + 8 & 15]) ^ w[20 + 1 + 2 & 15]) ^ w[20 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 ^ v0) ^ v1)) + 0x6ed9eba1) + (w[20 + 2 & 15] = (((((w[20 + 2 + 13 & 15] ^ w[20 + 2 + 8 & 15]) ^ w[20 + 2 + 2 & 15]) ^ w[20 + 2 & 15]) << 1) | ((((w[20 + 2 + 13 & 15] ^ w[20 + 2 + 8 & 15]) ^ w[20 + 2 + 2 & 15]) ^ w[20 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 ^ v4) ^ v0)) + 0x6ed9eba1) + (w[20 + 3 & 15] = (((((w[20 + 3 + 13 & 15] ^ w[20 + 3 + 8 & 15]) ^ w[20 + 3 + 2 & 15]) ^ w[20 + 3 & 15]) << 1) | ((((w[20 + 3 + 13 & 15] ^ w[20 + 3 + 8 & 15]) ^ w[20 + 3 + 2 & 15]) ^ w[20 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 ^ v3) ^ v4)) + 0x6ed9eba1) + (w[20 + 4 & 15] = (((((w[20 + 4 + 13 & 15] ^ w[20 + 4 + 8 & 15]) ^ w[20 + 4 + 2 & 15]) ^ w[20 + 4 & 15]) << 1) | ((((w[20 + 4 + 13 & 15] ^ w[20 + 4 + 8 & 15]) ^ w[20 + 4 + 2 & 15]) ^ w[20 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 ^ v2) ^ v3)) + 0x6ed9eba1) + (w[25 & 15] = (((((w[25 + 13 & 15] ^ w[25 + 8 & 15]) ^ w[25 + 2 & 15]) ^ w[25 & 15]) << 1) | ((((w[25 + 13 & 15] ^ w[25 + 8 & 15]) ^ w[25 + 2 & 15]) ^ w[25 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 ^ v1) ^ v2)) + 0x6ed9eba1) + (w[25 + 1 & 15] = (((((w[25 + 1 + 13 & 15] ^ w[25 + 1 + 8 & 15]) ^ w[25 + 1 + 2 & 15]) ^ w[25 + 1 & 15]) << 1) | ((((w[25 + 1 + 13 & 15] ^ w[25 + 1 + 8 & 15]) ^ w[25 + 1 + 2 & 15]) ^ w[25 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 ^ v0) ^ v1)) + 0x6ed9eba1) + (w[25 + 2 & 15] = (((((w[25 + 2 + 13 & 15] ^ w[25 + 2 + 8 & 15]) ^ w[25 + 2 + 2 & 15]) ^ w[25 + 2 & 15]) << 1) | ((((w[25 + 2 + 13 & 15] ^ w[25 + 2 + 8 & 15]) ^ w[25 + 2 + 2 & 15]) ^ w[25 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 ^ v4) ^ v0)) + 0x6ed9eba1) + (w[25 + 3 & 15] = (((((w[25 + 3 + 13 & 15] ^ w[25 + 3 + 8 & 15]) ^ w[25 + 3 + 2 & 15]) ^ w[25 + 3 & 15]) << 1) | ((((w[25 + 3 + 13 & 15] ^ w[25 + 3 + 8 & 15]) ^ w[25 + 3 + 2 & 15]) ^ w[25 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 ^ v3) ^ v4)) + 0x6ed9eba1) + (w[25 + 4 & 15] = (((((w[25 + 4 + 13 & 15] ^ w[25 + 4 + 8 & 15]) ^ w[25 + 4 + 2 & 15]) ^ w[25 + 4 & 15]) << 1) | ((((w[25 + 4 + 13 & 15] ^ w[25 + 4 + 8 & 15]) ^ w[25 + 4 + 2 & 15]) ^ w[25 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 ^ v2) ^ v3)) + 0x6ed9eba1) + (w[30 & 15] = (((((w[30 + 13 & 15] ^ w[30 + 8 & 15]) ^ w[30 + 2 & 15]) ^ w[30 & 15]) << 1) | ((((w[30 + 13 & 15] ^ w[30 + 8 & 15]) ^ w[30 + 2 & 15]) ^ w[30 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 ^ v1) ^ v2)) + 0x6ed9eba1) + (w[30 + 1 & 15] = (((((w[30 + 1 + 13 & 15] ^ w[30 + 1 + 8 & 15]) ^ w[30 + 1 + 2 & 15]) ^ w[30 + 1 & 15]) << 1) | ((((w[30 + 1 + 13 & 15] ^ w[30 + 1 + 8 & 15]) ^ w[30 + 1 + 2 & 15]) ^ w[30 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 ^ v0) ^ v1)) + 0x6ed9eba1) + (w[30 + 2 & 15] = (((((w[30 + 2 + 13 & 15] ^ w[30 + 2 + 8 & 15]) ^ w[30 + 2 + 2 & 15]) ^ w[30 + 2 & 15]) << 1) | ((((w[30 + 2 + 13 & 15] ^ w[30 + 2 + 8 & 15]) ^ w[30 + 2 + 2 & 15]) ^ w[30 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 ^ v4) ^ v0)) + 0x6ed9eba1) + (w[30 + 3 & 15] = (((((w[30 + 3 + 13 & 15] ^ w[30 + 3 + 8 & 15]) ^ w[30 + 3 + 2 & 15]) ^ w[30 + 3 & 15]) << 1) | ((((w[30 + 3 + 13 & 15] ^ w[30 + 3 + 8 & 15]) ^ w[30 + 3 + 2 & 15]) ^ w[30 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 ^ v3) ^ v4)) + 0x6ed9eba1) + (w[30 + 4 & 15] = (((((w[30 + 4 + 13 & 15] ^ w[30 + 4 + 8 & 15]) ^ w[30 + 4 + 2 & 15]) ^ w[30 + 4 & 15]) << 1) | ((((w[30 + 4 + 13 & 15] ^ w[30 + 4 + 8 & 15]) ^ w[30 + 4 + 2 & 15]) ^ w[30 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 ^ v2) ^ v3)) + 0x6ed9eba1) + (w[35 & 15] = (((((w[35 + 13 & 15] ^ w[35 + 8 & 15]) ^ w[35 + 2 & 15]) ^ w[35 & 15]) << 1) | ((((w[35 + 13 & 15] ^ w[35 + 8 & 15]) ^ w[35 + 2 & 15]) ^ w[35 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 ^ v1) ^ v2)) + 0x6ed9eba1) + (w[35 + 1 & 15] = (((((w[35 + 1 + 13 & 15] ^ w[35 + 1 + 8 & 15]) ^ w[35 + 1 + 2 & 15]) ^ w[35 + 1 & 15]) << 1) | ((((w[35 + 1 + 13 & 15] ^ w[35 + 1 + 8 & 15]) ^ w[35 + 1 + 2 & 15]) ^ w[35 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 ^ v0) ^ v1)) + 0x6ed9eba1) + (w[35 + 2 & 15] = (((((w[35 + 2 + 13 & 15] ^ w[35 + 2 + 8 & 15]) ^ w[35 + 2 + 2 & 15]) ^ w[35 + 2 & 15]) << 1) | ((((w[35 + 2 + 13 & 15] ^ w[35 + 2 + 8 & 15]) ^ w[35 + 2 + 2 & 15]) ^ w[35 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 ^ v4) ^ v0)) + 0x6ed9eba1) + (w[35 + 3 & 15] = (((((w[35 + 3 + 13 & 15] ^ w[35 + 3 + 8 & 15]) ^ w[35 + 3 + 2 & 15]) ^ w[35 + 3 & 15]) << 1) | ((((w[35 + 3 + 13 & 15] ^ w[35 + 3 + 8 & 15]) ^ w[35 + 3 + 2 & 15]) ^ w[35 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 ^ v3) ^ v4)) + 0x6ed9eba1) + (w[35 + 4 & 15] = (((((w[35 + 4 + 13 & 15] ^ w[35 + 4 + 8 & 15]) ^ w[35 + 4 + 2 & 15]) ^ w[35 + 4 & 15]) << 1) | ((((w[35 + 4 + 13 & 15] ^ w[35 + 4 + 8 & 15]) ^ w[35 + 4 + 2 & 15]) ^ w[35 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 & v2) | (v3 & (v1 ^ v2)))) + 0x8f1bbcdc) + (w[40 & 15] = (((((w[40 + 13 & 15] ^ w[40 + 8 & 15]) ^ w[40 + 2 & 15]) ^ w[40 & 15]) << 1) | ((((w[40 + 13 & 15] ^ w[40 + 8 & 15]) ^ w[40 + 2 & 15]) ^ w[40 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 & v1) | (v2 & (v0 ^ v1)))) + 0x8f1bbcdc) + (w[40 + 1 & 15] = (((((w[40 + 1 + 13 & 15] ^ w[40 + 1 + 8 & 15]) ^ w[40 + 1 + 2 & 15]) ^ w[40 + 1 & 15]) << 1) | ((((w[40 + 1 + 13 & 15] ^ w[40 + 1 + 8 & 15]) ^ w[40 + 1 + 2 & 15]) ^ w[40 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 & v0) | (v1 & (v4 ^ v0)))) + 0x8f1bbcdc) + (w[40 + 2 & 15] = (((((w[40 + 2 + 13 & 15] ^ w[40 + 2 + 8 & 15]) ^ w[40 + 2 + 2 & 15]) ^ w[40 + 2 & 15]) << 1) | ((((w[40 + 2 + 13 & 15] ^ w[40 + 2 + 8 & 15]) ^ w[40 + 2 + 2 & 15]) ^ w[40 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 & v4) | (v0 & (v3 ^ v4)))) + 0x8f1bbcdc) + (w[40 + 3 & 15] = (((((w[40 + 3 + 13 & 15] ^ w[40 + 3 + 8 & 15]) ^ w[40 + 3 + 2 & 15]) ^ w[40 + 3 & 15]) << 1) | ((((w[40 + 3 + 13 & 15] ^ w[40 + 3 + 8 & 15]) ^ w[40 + 3 + 2 & 15]) ^ w[40 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 & v3) | (v4 & (v2 ^ v3)))) + 0x8f1bbcdc) + (w[40 + 4 & 15] = (((((w[40 + 4 + 13 & 15] ^ w[40 + 4 + 8 & 15]) ^ w[40 + 4 + 2 & 15]) ^ w[40 + 4 & 15]) << 1) | ((((w[40 + 4 + 13 & 15] ^ w[40 + 4 + 8 & 15]) ^ w[40 + 4 + 2 & 15]) ^ w[40 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 & v2) | (v3 & (v1 ^ v2)))) + 0x8f1bbcdc) + (w[45 & 15] = (((((w[45 + 13 & 15] ^ w[45 + 8 & 15]) ^ w[45 + 2 & 15]) ^ w[45 & 15]) << 1) | ((((w[45 + 13 & 15] ^ w[45 + 8 & 15]) ^ w[45 + 2 & 15]) ^ w[45 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 & v1) | (v2 & (v0 ^ v1)))) + 0x8f1bbcdc) + (w[45 + 1 & 15] = (((((w[45 + 1 + 13 & 15] ^ w[45 + 1 + 8 & 15]) ^ w[45 + 1 + 2 & 15]) ^ w[45 + 1 & 15]) << 1) | ((((w[45 + 1 + 13 & 15] ^ w[45 + 1 + 8 & 15]) ^ w[45 + 1 + 2 & 15]) ^ w[45 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 & v0) | (v1 & (v4 ^ v0)))) + 0x8f1bbcdc) + (w[45 + 2 & 15] = (((((w[45 + 2 + 13 & 15] ^ w[45 + 2 + 8 & 15]) ^ w[45 + 2 + 2 & 15]) ^ w[45 + 2 & 15]) << 1) | ((((w[45 + 2 + 13 & 15] ^ w[45 + 2 + 8 & 15]) ^ w[45 + 2 + 2 & 15]) ^ w[45 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 & v4) | (v0 & (v3 ^ v4)))) + 0x8f1bbcdc) + (w[45 + 3 & 15] = (((((w[45 + 3 + 13 & 15] ^ w[45 + 3 + 8 & 15]) ^ w[45 + 3 + 2 & 15]) ^ w[45 + 3 & 15]) << 1) | ((((w[45 + 3 + 13 & 15] ^ w[45 + 3 + 8 & 15]) ^ w[45 + 3 + 2 & 15]) ^ w[45 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 & v3) | (v4 & (v2 ^ v3)))) + 0x8f1bbcdc) + (w[45 + 4 & 15] = (((((w[45 + 4 + 13 & 15] ^ w[45 + 4 + 8 & 15]) ^ w[45 + 4 + 2 & 15]) ^ w[45 + 4 & 15]) << 1) | ((((w[45 + 4 + 13 & 15] ^ w[45 + 4 + 8 & 15]) ^ w[45 + 4 + 2 & 15]) ^ w[45 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 & v2) | (v3 & (v1 ^ v2)))) + 0x8f1bbcdc) + (w[50 & 15] = (((((w[50 + 13 & 15] ^ w[50 + 8 & 15]) ^ w[50 + 2 & 15]) ^ w[50 & 15]) << 1) | ((((w[50 + 13 & 15] ^ w[50 + 8 & 15]) ^ w[50 + 2 & 15]) ^ w[50 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 & v1) | (v2 & (v0 ^ v1)))) + 0x8f1bbcdc) + (w[50 + 1 & 15] = (((((w[50 + 1 + 13 & 15] ^ w[50 + 1 + 8 & 15]) ^ w[50 + 1 + 2 & 15]) ^ w[50 + 1 & 15]) << 1) | ((((w[50 + 1 + 13 & 15] ^ w[50 + 1 + 8 & 15]) ^ w[50 + 1 + 2 & 15]) ^ w[50 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 & v0) | (v1 & (v4 ^ v0)))) + 0x8f1bbcdc) + (w[50 + 2 & 15] = (((((w[50 + 2 + 13 & 15] ^ w[50 + 2 + 8 & 15]) ^ w[50 + 2 + 2 & 15]) ^ w[50 + 2 & 15]) << 1) | ((((w[50 + 2 + 13 & 15] ^ w[50 + 2 + 8 & 15]) ^ w[50 + 2 + 2 & 15]) ^ w[50 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 & v4) | (v0 & (v3 ^ v4)))) + 0x8f1bbcdc) + (w[50 + 3 & 15] = (((((w[50 + 3 + 13 & 15] ^ w[50 + 3 + 8 & 15]) ^ w[50 + 3 + 2 & 15]) ^ w[50 + 3 & 15]) << 1) | ((((w[50 + 3 + 13 & 15] ^ w[50 + 3 + 8 & 15]) ^ w[50 + 3 + 2 & 15]) ^ w[50 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 & v3) | (v4 & (v2 ^ v3)))) + 0x8f1bbcdc) + (w[50 + 4 & 15] = (((((w[50 + 4 + 13 & 15] ^ w[50 + 4 + 8 & 15]) ^ w[50 + 4 + 2 & 15]) ^ w[50 + 4 & 15]) << 1) | ((((w[50 + 4 + 13 & 15] ^ w[50 + 4 + 8 & 15]) ^ w[50 + 4 + 2 & 15]) ^ w[50 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 & v2) | (v3 & (v1 ^ v2)))) + 0x8f1bbcdc) + (w[55 & 15] = (((((w[55 + 13 & 15] ^ w[55 + 8 & 15]) ^ w[55 + 2 & 15]) ^ w[55 & 15]) << 1) | ((((w[55 + 13 & 15] ^ w[55 + 8 & 15]) ^ w[55 + 2 & 15]) ^ w[55 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 & v1) | (v2 & (v0 ^ v1)))) + 0x8f1bbcdc) + (w[55 + 1 & 15] = (((((w[55 + 1 + 13 & 15] ^ w[55 + 1 + 8 & 15]) ^ w[55 + 1 + 2 & 15]) ^ w[55 + 1 & 15]) << 1) | ((((w[55 + 1 + 13 & 15] ^ w[55 + 1 + 8 & 15]) ^ w[55 + 1 + 2 & 15]) ^ w[55 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 & v0) | (v1 & (v4 ^ v0)))) + 0x8f1bbcdc) + (w[55 + 2 & 15] = (((((w[55 + 2 + 13 & 15] ^ w[55 + 2 + 8 & 15]) ^ w[55 + 2 + 2 & 15]) ^ w[55 + 2 & 15]) << 1) | ((((w[55 + 2 + 13 & 15] ^ w[55 + 2 + 8 & 15]) ^ w[55 + 2 + 2 & 15]) ^ w[55 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 & v4) | (v0 & (v3 ^ v4)))) + 0x8f1bbcdc) + (w[55 + 3 & 15] = (((((w[55 + 3 + 13 & 15] ^ w[55 + 3 + 8 & 15]) ^ w[55 + 3 + 2 & 15]) ^ w[55 + 3 & 15]) << 1) | ((((w[55 + 3 + 13 & 15] ^ w[55 + 3 + 8 & 15]) ^ w[55 + 3 + 2 & 15]) ^ w[55 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 & v3) | (v4 & (v2 ^ v3)))) + 0x8f1bbcdc) + (w[55 + 4 & 15] = (((((w[55 + 4 + 13 & 15] ^ w[55 + 4 + 8 & 15]) ^ w[55 + 4 + 2 & 15]) ^ w[55 + 4 & 15]) << 1) | ((((w[55 + 4 + 13 & 15] ^ w[55 + 4 + 8 & 15]) ^ w[55 + 4 + 2 & 15]) ^ w[55 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 ^ v2) ^ v3)) + 0xca62c1d6) + (w[60 & 15] = (((((w[60 + 13 & 15] ^ w[60 + 8 & 15]) ^ w[60 + 2 & 15]) ^ w[60 & 15]) << 1) | ((((w[60 + 13 & 15] ^ w[60 + 8 & 15]) ^ w[60 + 2 & 15]) ^ w[60 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 ^ v1) ^ v2)) + 0xca62c1d6) + (w[60 + 1 & 15] = (((((w[60 + 1 + 13 & 15] ^ w[60 + 1 + 8 & 15]) ^ w[60 + 1 + 2 & 15]) ^ w[60 + 1 & 15]) << 1) | ((((w[60 + 1 + 13 & 15] ^ w[60 + 1 + 8 & 15]) ^ w[60 + 1 + 2 & 15]) ^ w[60 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 ^ v0) ^ v1)) + 0xca62c1d6) + (w[60 + 2 & 15] = (((((w[60 + 2 + 13 & 15] ^ w[60 + 2 + 8 & 15]) ^ w[60 + 2 + 2 & 15]) ^ w[60 + 2 & 15]) << 1) | ((((w[60 + 2 + 13 & 15] ^ w[60 + 2 + 8 & 15]) ^ w[60 + 2 + 2 & 15]) ^ w[60 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 ^ v4) ^ v0)) + 0xca62c1d6) + (w[60 + 3 & 15] = (((((w[60 + 3 + 13 & 15] ^ w[60 + 3 + 8 & 15]) ^ w[60 + 3 + 2 & 15]) ^ w[60 + 3 & 15]) << 1) | ((((w[60 + 3 + 13 & 15] ^ w[60 + 3 + 8 & 15]) ^ w[60 + 3 + 2 & 15]) ^ w[60 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 ^ v3) ^ v4)) + 0xca62c1d6) + (w[60 + 4 & 15] = (((((w[60 + 4 + 13 & 15] ^ w[60 + 4 + 8 & 15]) ^ w[60 + 4 + 2 & 15]) ^ w[60 + 4 & 15]) << 1) | ((((w[60 + 4 + 13 & 15] ^ w[60 + 4 + 8 & 15]) ^ w[60 + 4 + 2 & 15]) ^ w[60 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 ^ v2) ^ v3)) + 0xca62c1d6) + (w[65 & 15] = (((((w[65 + 13 & 15] ^ w[65 + 8 & 15]) ^ w[65 + 2 & 15]) ^ w[65 & 15]) << 1) | ((((w[65 + 13 & 15] ^ w[65 + 8 & 15]) ^ w[65 + 2 & 15]) ^ w[65 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 ^ v1) ^ v2)) + 0xca62c1d6) + (w[65 + 1 & 15] = (((((w[65 + 1 + 13 & 15] ^ w[65 + 1 + 8 & 15]) ^ w[65 + 1 + 2 & 15]) ^ w[65 + 1 & 15]) << 1) | ((((w[65 + 1 + 13 & 15] ^ w[65 + 1 + 8 & 15]) ^ w[65 + 1 + 2 & 15]) ^ w[65 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 ^ v0) ^ v1)) + 0xca62c1d6) + (w[65 + 2 & 15] = (((((w[65 + 2 + 13 & 15] ^ w[65 + 2 + 8 & 15]) ^ w[65 + 2 + 2 & 15]) ^ w[65 + 2 & 15]) << 1) | ((((w[65 + 2 + 13 & 15] ^ w[65 + 2 + 8 & 15]) ^ w[65 + 2 + 2 & 15]) ^ w[65 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 ^ v4) ^ v0)) + 0xca62c1d6) + (w[65 + 3 & 15] = (((((w[65 + 3 + 13 & 15] ^ w[65 + 3 + 8 & 15]) ^ w[65 + 3 + 2 & 15]) ^ w[65 + 3 & 15]) << 1) | ((((w[65 + 3 + 13 & 15] ^ w[65 + 3 + 8 & 15]) ^ w[65 + 3 + 2 & 15]) ^ w[65 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 ^ v3) ^ v4)) + 0xca62c1d6) + (w[65 + 4 & 15] = (((((w[65 + 4 + 13 & 15] ^ w[65 + 4 + 8 & 15]) ^ w[65 + 4 + 2 & 15]) ^ w[65 + 4 & 15]) << 1) | ((((w[65 + 4 + 13 & 15] ^ w[65 + 4 + 8 & 15]) ^ w[65 + 4 + 2 & 15]) ^ w[65 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 ^ v2) ^ v3)) + 0xca62c1d6) + (w[70 & 15] = (((((w[70 + 13 & 15] ^ w[70 + 8 & 15]) ^ w[70 + 2 & 15]) ^ w[70 & 15]) << 1) | ((((w[70 + 13 & 15] ^ w[70 + 8 & 15]) ^ w[70 + 2 & 15]) ^ w[70 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 ^ v1) ^ v2)) + 0xca62c1d6) + (w[70 + 1 & 15] = (((((w[70 + 1 + 13 & 15] ^ w[70 + 1 + 8 & 15]) ^ w[70 + 1 + 2 & 15]) ^ w[70 + 1 & 15]) << 1) | ((((w[70 + 1 + 13 & 15] ^ w[70 + 1 + 8 & 15]) ^ w[70 + 1 + 2 & 15]) ^ w[70 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 ^ v0) ^ v1)) + 0xca62c1d6) + (w[70 + 2 & 15] = (((((w[70 + 2 + 13 & 15] ^ w[70 + 2 + 8 & 15]) ^ w[70 + 2 + 2 & 15]) ^ w[70 + 2 & 15]) << 1) | ((((w[70 + 2 + 13 & 15] ^ w[70 + 2 + 8 & 15]) ^ w[70 + 2 + 2 & 15]) ^ w[70 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 ^ v4) ^ v0)) + 0xca62c1d6) + (w[70 + 3 & 15] = (((((w[70 + 3 + 13 & 15] ^ w[70 + 3 + 8 & 15]) ^ w[70 + 3 + 2 & 15]) ^ w[70 + 3 & 15]) << 1) | ((((w[70 + 3 + 13 & 15] ^ w[70 + 3 + 8 & 15]) ^ w[70 + 3 + 2 & 15]) ^ w[70 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 ^ v3) ^ v4)) + 0xca62c1d6) + (w[70 + 4 & 15] = (((((w[70 + 4 + 13 & 15] ^ w[70 + 4 + 8 & 15]) ^ w[70 + 4 + 2 & 15]) ^ w[70 + 4 & 15]) << 1) | ((((w[70 + 4 + 13 & 15] ^ w[70 + 4 + 8 & 15]) ^ w[70 + 4 + 2 & 15]) ^ w[70 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
  v4 += (((((v0 >> 27) | (v0 << 32 - 27)) + ((v1 ^ v2) ^ v3)) + 0xca62c1d6) + (w[75 & 15] = (((((w[75 + 13 & 15] ^ w[75 + 8 & 15]) ^ w[75 + 2 & 15]) ^ w[75 & 15]) << 1) | ((((w[75 + 13 & 15] ^ w[75 + 8 & 15]) ^ w[75 + 2 & 15]) ^ w[75 & 15]) >> 32 - 1))));
  v1 = ((v1 >> 2) | (v1 << 32 - 2));
  v3 += (((((v4 >> 27) | (v4 << 32 - 27)) + ((v0 ^ v1) ^ v2)) + 0xca62c1d6) + (w[75 + 1 & 15] = (((((w[75 + 1 + 13 & 15] ^ w[75 + 1 + 8 & 15]) ^ w[75 + 1 + 2 & 15]) ^ w[75 + 1 & 15]) << 1) | ((((w[75 + 1 + 13 & 15] ^ w[75 + 1 + 8 & 15]) ^ w[75 + 1 + 2 & 15]) ^ w[75 + 1 & 15]) >> 32 - 1))));
  v0 = ((v0 >> 2) | (v0 << 32 - 2));
  v2 += (((((v3 >> 27) | (v3 << 32 - 27)) + ((v4 ^ v0) ^ v1)) + 0xca62c1d6) + (w[75 + 2 & 15] = (((((w[75 + 2 + 13 & 15] ^ w[75 + 2 + 8 & 15]) ^ w[75 + 2 + 2 & 15]) ^ w[75 + 2 & 15]) << 1) | ((((w[75 + 2 + 13 & 15] ^ w[75 + 2 + 8 & 15]) ^ w[75 + 2 + 2 & 15]) ^ w[75 + 2 & 15]) >> 32 - 1))));
  v4 = ((v4 >> 2) | (v4 << 32 - 2));
  v1 += (((((v2 >> 27) | (v2 << 32 - 27)) + ((v3 ^ v4) ^ v0)) + 0xca62c1d6) + (w[75 + 3 & 15] = (((((w[75 + 3 + 13 & 15] ^ w[75 + 3 + 8 & 15]) ^ w[75 + 3 + 2 & 15]) ^ w[75 + 3 & 15]) << 1) | ((((w[75 + 3 + 13 & 15] ^ w[75 + 3 + 8 & 15]) ^ w[75 + 3 + 2 & 15]) ^ w[75 + 3 & 15]) >> 32 - 1))));
  v3 = ((v3 >> 2) | (v3 << 32 - 2));
  v0 += (((((v1 >> 27) | (v1 << 32 - 27)) + ((v2 ^ v3) ^ v4)) + 0xca62c1d6) + (w[75 + 4 & 15] = (((((w[75 + 4 + 13 & 15] ^ w[75 + 4 + 8 & 15]) ^ w[75 + 4 + 2 & 15]) ^ w[75 + 4 & 15]) << 1) | ((((w[75 + 4 + 13 & 15] ^ w[75 + 4 + 8 & 15]) ^ w[75 + 4 + 2 & 15]) ^ w[75 + 4 & 15]) >> 32 - 1))));
  v2 = ((v2 >> 2) | (v2 << 32 - 2));
#ifdef ARRAY
#else
  (ctx -> sha1_ctx_s::hash)[0] += v0;
  (ctx -> sha1_ctx_s::hash)[1] += v1;
  (ctx -> sha1_ctx_s::hash)[2] += v2;
  (ctx -> sha1_ctx_s::hash)[3] += v3;
  (ctx -> sha1_ctx_s::hash)[4] += v4;
#endif
}
 }

extern "C" { void sha1_begin(sha1_ctx ctx[1UL])
{
  (ctx -> sha1_ctx_s::count)[0] = ((ctx -> sha1_ctx_s::count)[1] = 0);
  (ctx -> sha1_ctx_s::hash)[0] = 0x67452301;
  (ctx -> sha1_ctx_s::hash)[1] = 0xefcdab89;
  (ctx -> sha1_ctx_s::hash)[2] = 0x98badcfe;
  (ctx -> sha1_ctx_s::hash)[3] = 0x10325476;
  (ctx -> sha1_ctx_s::hash)[4] = 0xc3d2e1f0;
}
 }
/* SHA1 hash data in an array of bytes into hash buffer and */
/* call the hash_compile function as required.              */

extern "C" { void sha1_hash(const unsigned char data[],unsigned long len,sha1_ctx ctx[1UL])
{
  uint_32t pos = ((ctx -> sha1_ctx_s::count)[0] & (64 - 1));
  uint_32t space = (64 - pos);
  const unsigned char *sp = data;
  if (((ctx -> sha1_ctx_s::count)[0] += len) < len) 
    ++(ctx -> sha1_ctx_s::count)[1];
/* tranfer whole blocks if possible  */
  while(len >= space){
    memcpy((((unsigned char *)(ctx -> sha1_ctx_s::wbuf)) + pos),sp,space);
    sp += space;
    len -= space;
    space = 64;
    pos = 0;
{
      int _i = 64 >> 2;
      while((_i--))
        ((uint_32t *)(&ctx -> sha1_ctx_s::wbuf))[_i] = ((
{
          register unsigned int __v;
          register unsigned int __x = ((uint_32t *)(&ctx -> sha1_ctx_s::wbuf))[_i];
          if (0) 
            __v = (((((__x & 0xff000000) >> 24) | ((__x & 0x00ff0000) >> 8)) | ((__x & 0x0000ff00) << 8)) | ((__x & 0xFF) << 24));
          else 
            asm ("bswap %0" : "=r" (__v) : "0" (__x));
          __v;
        }));
    };
    sha1_compile(ctx);
  }
  memcpy((((unsigned char *)(ctx -> sha1_ctx_s::wbuf)) + pos),sp,len);
}
 }
/* SHA1 final padding and digest calculation  */

extern "C" { void sha1_end(unsigned char hval[],sha1_ctx ctx[1UL])
{
  uint_32t i = ((ctx -> sha1_ctx_s::count)[0] & (64 - 1));
/* put bytes in the buffer in an order in which references to   */
/* 32-bit words will put bytes with lower addresses into the    */
/* top of 32 bit words on BOTH big and little endian machines   */
{
    int _i = ((i + 3) >> 2);
    while((_i--))
      ((uint_32t *)(&ctx -> sha1_ctx_s::wbuf))[_i] = ((
{
        register unsigned int __v;
        register unsigned int __x = ((uint_32t *)(&ctx -> sha1_ctx_s::wbuf))[_i];
        if (0) 
          __v = (((((__x & 0xff000000) >> 24) | ((__x & 0x00ff0000) >> 8)) | ((__x & 0x0000ff00) << 8)) | ((__x & 0xFF) << 24));
        else 
          asm ("bswap %0" : "=r" (__v) : "0" (__x));
        __v;
      }));
  };
/* we now need to mask valid bytes and add the padding which is */
/* a single 1 bit and as many zero bits as necessary. Note that */
/* we can always add the first padding byte here because the    */
/* buffer always has at least one empty slot                    */
  (ctx -> sha1_ctx_s::wbuf)[i >> 2] &= (0xffffff80 << (8 * (~i & 3)));
  (ctx -> sha1_ctx_s::wbuf)[i >> 2] |= (0x00000080 << (8 * (~i & 3)));
/* we need 9 or more empty positions, one for the padding byte  */
/* (above) and eight for the length count. If there is not      */
/* enough space, pad and empty the buffer                       */
  if (i > (64 - 9)) {
    if (i < 60) 
      (ctx -> sha1_ctx_s::wbuf)[15] = 0;
    sha1_compile(ctx);
    i = 0;
  }
  else 
/* compute a word index for the empty buffer positions  */
    i = ((i >> 2) + 1);
/* and zero pad all but last two positions        */
  while(i < 14)
    (ctx -> sha1_ctx_s::wbuf)[i++] = 0;
/* the following 32-bit length fields are assembled in the      */
/* wrong byte order on little endian machines but this is       */
/* corrected later since they are only ever used as 32-bit      */
/* word values.                                                 */
  (ctx -> sha1_ctx_s::wbuf)[14] = (((ctx -> sha1_ctx_s::count)[1] << 3) | ((ctx -> sha1_ctx_s::count)[0] >> 29));
  (ctx -> sha1_ctx_s::wbuf)[15] = ((ctx -> sha1_ctx_s::count)[0] << 3);
  sha1_compile(ctx);
/* extract the hash value as bytes in case the hash buffer is   */
/* misaligned for 32-bit words                                  */
  for (i = 0; i < 20; ++i) 
    hval[i] = ((unsigned char )((ctx -> sha1_ctx_s::hash)[i >> 2] >> (8 * (~i & 3))));
}
 }

extern "C" { void sha1(unsigned char hval[],const unsigned char data[],unsigned long len)
{
  sha1_ctx cx[1UL];
  sha1_begin(cx);
  sha1_hash(data,len,cx);
  sha1_end(hval,cx);
}
 }
#if defined(__cplusplus)
}
#endif
/////////////////////////////////////////////////////////////////
////////////////////////////////#include "uts.h"
#ifndef _UTS_H
#define _UTS_H
#define UTS_VERSION "2.1"
/***********************************************************
 *  Tree node descriptor and statistics                    *
 ***********************************************************/
#define MAXNUMCHILDREN    100  // cap on children (BIN root is exempt)

struct node_t 
{
// depth of this node in the tree
  int height;
// number of children, -1 => not yet determined
  int numChildren;
/* for RNG state associated with this node */
  struct state_t state;
}
;
typedef struct node_t Node;
/* Tree  parameters */
extern double b_0;
extern int rootId;
extern int nonLeafBF;
extern double nonLeafProb;
/* Benchmark parameters */
extern int computeGranularity;
extern int debug;
extern int verbose;
/* Utility Functions */
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))
void serTreeSearch_block_bf(int depth,unsigned long long *subtreesize,_BlockStack *_stack,int _depth);
void serTreeSearch_block_df(int depth,unsigned long long *subtreesize,_BlockStack *_stack,int _depth,int i);
void serTreeSearch(int depth,Node *parent,int numChildren,unsigned long long *subtreesize);
int uts_paramsToStr(char *strBuf,int ind);
void uts_read_file(char *file);
void uts_print_params();
/* Common tree routines */
void uts_initRoot(Node *root);
int uts_numChildren(Node *parent);
int uts_numChildren_bin(Node *parent);
int uts_numChildren_geo(Node *parent);
int uts_childType(Node *parent);
void uts_show_stats();
int uts_check_result();
#endif /* _UTS_H */
/////////////////////////////////////////////////////////////
///////////////////////////////////////////////////uts.cpp
int _expandSize = 2048;
int _minBlockSize = 8;
int _maxNumChildren_0 = 2000;
int _maxNumChildren_1 = 8;
int _maxNumChildren = 2000;
/***********************************************************
 *  Global state                                           *
 ***********************************************************/
/* forcing 8 bytes size in -m32 and -m64 */
unsigned long long bots_number_of_tasks = 0;
unsigned long long nLeaves = 0;
int maxTreeDepth = 0;
/***********************************************************
 * Tree generation strategy is controlled via various      *
 * parameters set from the command line.  The parameters   *
 * and their default values are given below.               *
 * Trees are generated using a Galton-Watson process, in   *
 * which the branching factor of each node is a random     *
 * variable.                                               *
 *                                                         *
 * The random variable follow a binomial distribution.     *
 ***********************************************************/
// default branching factor at the root
double b_0 = 4.0;
// default seed for RNG state at root
int rootId = 0;
/***********************************************************
 *  The branching factor at the root is specified by b_0.
 *  The branching factor below the root follows an
 *     identical binomial distribution at all nodes.
 *  A node has m children with prob q, or no children with
 *     prob (1-q).  The expected branching factor is q * m.
 *
 *  Default parameter values
 ***********************************************************/
// m
int nonLeafBF = 4;
// q
double nonLeafProb = 15.0 / 64.0;
/***********************************************************
 * compute granularity - number of rng evaluations per
 * tree node
 ***********************************************************/
int computeGranularity = 1;
/***********************************************************
 * expected results for execution
 ***********************************************************/
unsigned long long exp_tree_size = 0;
int exp_tree_depth = 0;
unsigned long long exp_num_leaves = 0;
/***********************************************************
 *  FUNCTIONS                                              *
 ***********************************************************/

void uts_initRoot(Node *root)
{
  root -> node_t::height = 0;
// means not yet determined
  root -> node_t::numChildren = -1;
  rng_init(root -> node_t::state.state_t::state,rootId);
}

int uts_numChildren(Node *node)
{
  int numChildren = 0;
// determine the number of children
  if ((node -> node_t::height) == 0) 
    numChildren = ((int )(floor(b_0)));
  else {
// distribution is identical everywhere below root
    int v = rng_rand(node -> node_t::state.state_t::state);
    double d = rng_toProb(v);
    numChildren = ((d < nonLeafProb)?nonLeafBF : 0);
  }
// limit number of children (only a BIN root can have more than MAXNUMCHILDREN)
  if ((node -> node_t::height) != 0) {
    if (numChildren > 100) {{
        fprintf(stdout,"*** Number of children truncated from %d to %d\n",numChildren,100);
      };
      numChildren = 100;
    }
  }
/* including info into node */
  node -> node_t::numChildren = numChildren;
  return numChildren;
}
/***********************************************************
 * Recursive depth-first implementation                    *
 ***********************************************************/

unsigned long long serial_uts(Node *root)
{
  unsigned long long num_nodes = 0;
{
    fprintf(stdout,"Computing Unbalance Tree Search algorithm ");
  };
  _minBlockSize = _maxNumChildren_1;
  int depth = 0;
  int numChildren = uts_numChildren(root);
  
#pragma recursionCallEntry
  serTreeSearch(depth,root,numChildren,&num_nodes);
{
    fprintf(stdout," completed!\n");
  };
  return num_nodes;
}

void serTreeSearch(int depth,struct node_t *parent,int numChildren,unsigned long long *subtreesize)
{
  for (int i = 0; i < ((depth?_maxNumChildren_1 : _maxNumChildren_0)); i++) {
    if (i == 0) 
       *subtreesize += 1;
    if (i >= numChildren) 
      return ;
    struct node_t *n = (struct node_t *)(__builtin_alloca(sizeof(struct node_t )));
    n -> node_t::height = ((parent -> node_t::height) + 1);
    for (int j = 0; j < computeGranularity; j++) {
      rng_spawn(parent -> node_t::state.state_t::state,n -> node_t::state.state_t::state,i);
    }
    int next_numChildren = uts_numChildren(n);
    serTreeSearch((depth + 1),n,next_numChildren,subtreesize);
  }
}

void uts_read_file(char *filename)
{
  FILE *fin;
  if ((fin = fopen(filename,"r")) == 0L) {{
      fprintf(stdout,"Could not open input file (%s)\n",filename);
    };
    exit(-1);
  }
  fscanf(fin,"%lf %lf %d %d %d %llu %d %llu",&b_0,&nonLeafProb,&nonLeafBF,&rootId,&computeGranularity,&exp_tree_size,&exp_tree_depth,&exp_num_leaves);
  fclose(fin);
  _maxNumChildren_0 = b_0;
  _maxNumChildren_1 = nonLeafBF;
  computeGranularity = ((1 > computeGranularity)?1 : computeGranularity);
// Printing input data
{
    fprintf(stdout,"\n");
  };
{
    fprintf(stdout,"Root branching factor                = %f\n",b_0);
  };
{
    fprintf(stdout,"Root seed (0 <= 2^31)                = %d\n",rootId);
  };
{
    fprintf(stdout,"Probability of non-leaf node         = %f\n",nonLeafProb);
  };
{
    fprintf(stdout,"Number of children for non-leaf node = %d\n",nonLeafBF);
  };
{
    fprintf(stdout,"E(n)                                 = %f\n",(nonLeafProb * nonLeafBF));
  };
{
    fprintf(stdout,"E(s)                                 = %f\n",(1.0 / (1.0 - (nonLeafProb * nonLeafBF))));
  };
{
    fprintf(stdout,"Compute granularity                  = %d\n",computeGranularity);
  };
{
    fprintf(stdout,"Random number generator              = ");
  };
  rng_showtype();
}

void uts_show_stats()
{
  int chunkSize = 0;
{
    fprintf(stdout,"\n");
  };
{
    fprintf(stdout,"Tree size                            = %llu\n",bots_number_of_tasks);
  };
{
    fprintf(stdout,"Maximum tree depth                   = %d\n",maxTreeDepth);
  };
{
    fprintf(stdout,"Chunk size                           = %d\n",chunkSize);
  };
{
    fprintf(stdout,"Number of leaves                     = %llu (%.2f%%)\n",nLeaves,((nLeaves / ((float )bots_number_of_tasks)) * 100.0));
  };
// bots_message("Wallclock time                       = %.3f sec\n", bots_time_program );
// bots_message("Overall performance                  = %.0f nodes/sec\n", (bots_number_of_tasks / bots_time_program) );
}

int uts_check_result()
{
  int answer = 1;
  if (bots_number_of_tasks != exp_tree_size) {
    answer = 0;
{
      fprintf(stdout,"Incorrect tree size result (%llu instead of %llu).\n",bots_number_of_tasks,exp_tree_size);
    };
  }
  return answer;
}
char bots_arg_file[255UL] = "";

int main(int argc,char **argv)
{
  if ((argc != 2) && (argc != 3)) {
    printf("Usage: uts [inputfile]\n");
    exit(0);
  }
  if (argc > 2) {
    _expandSize = atoi(argv[2]);
    printf("_expandSize is: %d\n",_expandSize);
//printf("_maxNumChildren is: %d\n", _maxNumChildren);
  }
  strcpy(bots_arg_file,argv[1]);
  Node root;
  uts_read_file(bots_arg_file);
  uts_initRoot(&root);
  bots_number_of_tasks = serial_uts(&root);
  uts_show_stats();
  if ((uts_check_result())) {
    printf("Successfully exit!\n");
  }
  return 0;
}
