/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#ifndef GLOBALS_H_
#define GLOBALS_H_

#include "rose.h"
#include <vector>
#include <string>
using namespace std;

extern bool g_useSplice;
extern bool g_useSOA;
extern int g_blockSize;
extern int g_spliceDepth;
extern bool g_useReexpand;

enum {
	Splice_GeneralOrder,
	Splice_NodeBasedOrder,
	Splice_NoOrder,
};

extern int g_spliceOptimizationLevel;

extern SgFunctionDeclaration *g_recursiveFunction;
extern SgMemberFunctionDeclaration *g_recursiveMemberFunction;

// call path in order of
// 0: call to recursive function
// g_numCallSites - 1: call in parallel for loop
// g_intermediaryFunctions has function declaration of each call site
// hence g_intermediaryFunctions[0] == g_recursiveFunction
extern vector<SgFunctionCallExp *> g_intermediaryCallSites;
extern vector<SgFunctionDeclaration *> g_intermediaryFunctions;
extern int g_numCallSites;


extern SgForStatement *g_enclosingForStmt;
extern SgForStatement *g_recursiveCallEnclosingForStmt;//For BF-DF toggling call code
extern SgGlobal *g_global;

extern SgInitializedName *g_nodeParam;
extern int g_nodeParamIdx;

class MethodInfo;
extern MethodInfo g_methodInfo;
extern SgPragmaDeclaration *g_afterClassDecsPragma;
extern vector<int> g_outerForLoopInvariantArgIdxs;

extern string g_recursiveTypeName;

extern SgPointerType *g_recursiveType; // this will be a pointer type
extern SgNamedType *g_blockType;
extern SgNamedType *g_blockSetType;
extern SgNamedType *g_blockStackType;
extern SgNamedType *g_intermediaryBlockType;
extern SgNamedType *g_intermediaryStateType;
extern SgNamedType *g_pointType;

extern SgClassDeclaration *g_recursiveClassDec;
extern SgClassDeclaration *g_blockClassDec;
extern SgClassDeclaration *g_blockStackClassDec;
extern SgClassDeclaration *g_blockSetClassDec;
extern SgClassDeclaration *g_intermediaryStateClassDec;
extern SgClassDeclaration *g_intermediaryBlockClassDec;
extern SgClassDeclaration *g_pointClassDec;

extern SgMemberFunctionSymbol *g_interBlockResetSym;
extern SgMemberFunctionSymbol *g_interBlockNextSym;
extern SgMemberFunctionSymbol *g_blockStackGetSym;
extern SgMemberFunctionSymbol *g_blockGetSym;
extern SgMemberFunctionSymbol *g_blockAddSym;
extern SgMemberFunctionSymbol *g_blockRecycleSym;
extern SgMemberFunctionSymbol *g_blockCopyUpSym;
extern SgMemberFunctionSymbol *g_blockCopyBackSym;
extern SgMemberFunctionSymbol *g_spliceNodeAddSym;

#define g_suffixMapRoot "_mapRoot"
#define g_suffixMapTree "_mapTree"
#define g_suffixBlock "_block_bf"
#define g_suffixBlock_DF "_block_df"
extern int g_num_recursiveCall;
extern vector<string> g_suffixBlock_DF_i;
//#define g_suffixBlock_DF0 "_block_df0"
//#define g_suffixBlock_DF1 "_block_df1"
#define g_suffixBlockSplice "_blockSplice"
#define g_suffixPrologue "_prologue"
#define g_suffixEpilogue "_epilogue"

#endif /* GLOBALS_H_ */
