/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#include "rose.h"

#include "IntermediaryStateClasses.h"
#include "Globals.h"
#include "Utils.h"
#include "CommonTransform.h"

#include <iostream>
#include <fstream>
using namespace std;

using namespace SageBuilder;
using namespace SageInterface;

void IntermediaryStateClasses::generate() {

	SgFunctionCallExp *call = g_intermediaryCallSites[g_numCallSites - 1];
	SgStatement *callStmt = getEnclosingStatement(call);
	SgBasicBlock *block = isSgBasicBlock(g_enclosingForStmt->get_loop_body());
	vector<SgStatement *> prologue, epilogue;
	vector<SgVariableDeclaration *> usedInEpilogueVarDecs;
	CommonTransform::proEpilogueAnalysis(block, callStmt, prologue, epilogue, usedInEpilogueVarDecs);

	//SageInterface::insertHeader("interstate.h", PreprocessingInfo::after, false, g_global);
	if (g_afterClassDecsPragma == NULL) {
		SageInterface::addTextForUnparser(g_global, "#include \"interstate.h\"\n", AstUnparseAttribute::e_before);
	} else {
		SageInterface::addTextForUnparser(g_afterClassDecsPragma, "#include \"interstate.h\"\n", AstUnparseAttribute::e_before);
	}

	ofstream file;
	file.open ("interstate.h");
	file << "#include <cassert>\n"
			<< getClassForwardDeclareString()
	<< "\n"
	<< "struct _IntermediaryState {\n";
	for (vector<SgVariableDeclaration *>::const_iterator i = usedInEpilogueVarDecs.begin();
			i != usedInEpilogueVarDecs.end(); i++) {
		SgVariableDeclaration *varDec = *i;
		SgInitializedName* initName = *(varDec->get_variables().begin());
		file << "\t" << getTypeString(initName->get_type(), true) << " " << string(initName->get_name()) << ";\n";
	}
	file << "};\n"
	<< "\n"
	<< "class _IntermediaryBlock {\n"
	<< "public:\n"
	<< "	_IntermediaryBlock();\n"
	<< "	~_IntermediaryBlock();\n"
	<< "\n"
	<< "	_IntermediaryState* next();\n"
	<< "	_IntermediaryState* get(int pos);\n"
	<< "	void reset();\n"
	<< "\n"
	<< "	_Block *block;\n"
	<< "	static int max_block;\n"
	<< "private:\n"
	<< "	int pos;\n"
	<< "	_IntermediaryState* data;\n"
	<< "};\n"
	<< "\n"
	<< "int _IntermediaryBlock::max_block = 128;\n"
	<< "\n"
	<< "_IntermediaryBlock::_IntermediaryBlock() {\n"
	<< "	pos = 0;\n"
	<< "	data = new _IntermediaryState[max_block];\n"
	<< "}\n"
	<< "\n"
	<< "_IntermediaryBlock::~_IntermediaryBlock() {\n"
	<< "	delete [] data;\n"
	<< "}\n"
	<< "\n"
	<< "_IntermediaryState* _IntermediaryBlock::next() {\n"
	<< "	assert(pos < max_block);\n"
	<< "	return &(data[pos++]);\n"
	<< "}\n"
	<< "\n"
	<< "_IntermediaryState* _IntermediaryBlock::get(int p) {\n"
	<< "	assert(p < pos);\n"
	<< "	return &(data[p]);\n"
	<< "}\n"
	<< "\n"
	<< "void _IntermediaryBlock::reset() {\n"
	<< "	pos = 0;\n"
	<< "}\n";

	file.close();
}
