/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#ifndef METHODINFO_H_
#define METHODINFO_H_

#include "rose.h"
#include <vector>
using namespace std;

class MethodInfo {
public:
	void analyze();

private:
	void generateMethodInfo();
	void setNodeParam();
	void analyzeRecursiveMethod();
	void findLoopVariantArgs();

    void findRecursiveVariant(vector<SgFunctionCallExp*> & allRecursiveCalls);
public:
	vector<int> recursiveMethodLoopVariantParamIdxs;
	vector<int> outerForLoopVariantArgIdxs;
	vector<int> recursiveMethodAllLoopVariantParamsIdxs;
	vector<int> recursiveMethodAdditionalLoopVariantParamsIdxs;

	vector<SgVarRefExp *> outerForLoopVariantArgs;
	vector<SgVarRefExp *> outerForLoopInvariantArgs;

	vector<SgInitializedName *> recursiveMethodLoopVariantParams;
	vector<SgInitializedName *> recursiveMethodLoopInvariantParams;

	// loop variant/invariant params, including/excluding recursive function local args that are loop variant
	vector<SgInitializedName *> recursiveMethodAllLoopVariantParams;
	vector<SgInitializedName *> recursiveMethodAllLoopInvariantParams;

	// recursive function local args that are loop variant
	vector<SgInitializedName *> recursiveMethodAdditionalLoopVariantParams;


	vector<SgInitializedName *> recursiveMethodAllLoopInvariantParamsIncludingNode;

	// var names of local loop variant args
	// not actually used now, but might be useful sometime
	// ex) in vptree, arg name: dist, param name: upperDist
	//   recursiveMethodAdditionalLoopVariantArgVarNames will contain [dist]
	set<string> recursiveMethodAdditionalLoopVariantArgVarNames;

	struct IntermediaryInfo {
		IntermediaryInfo(SgFunctionCallExp *c, SgFunctionDeclaration *f) {
			call = c;
			functionDec = f;
		}

		SgFunctionCallExp *call;
		SgFunctionDeclaration *functionDec;
		vector<int> loopVariantArgIdxs;
		vector<int> loopVariantParamIdxs;
		vector<SgVarRefExp *> loopVariantArgs;
		vector<SgVarRefExp *> loopInvariantArgs;
		vector<SgInitializedName *> loopVariantParams;
		vector<SgInitializedName *> loopInvariantParams;
	};

	vector<IntermediaryInfo *> intermediaryInfos;
};

#endif /* METHODINFO_H_ */
