/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#include "RecursiveCallSet.h"
#include "Utils.h"
#include "Globals.h"

using namespace SageBuilder;
using namespace SageInterface;

void RecursiveCallSet::analysis(SgBasicBlock *body, const vector<SgInitializedName *> &variantParams, string orgFuncName) {
	set<string> loopVariantArgsPropagation;
	for (int i = 0; i < variantParams.size(); i++) {
		SgInitializedName *param = variantParams[i];
		loopVariantArgsPropagation.insert(param->get_name());
	}
	vector<SgFunctionCallExp *> recursionPoints;
	getRecursionPoints(body, &recursionPoints, orgFuncName);
	vector<SgStatement *> recursionPointStmts;
	for (int i = 0; i < recursionPoints.size(); i++) {
		recursionPointStmts.push_back(getEnclosingStatement(recursionPoints[i]));
	}
	//cout << "recursionPointStmts.size(): " << recursionPointStmts.size() << endl;

	set<SgStatement *> loopVariantControlStmts;
	gatherLoopVariantStmts(&loopVariantControlStmts, &loopVariantArgsPropagation,
			body, recursionPointStmts);

	map<SgStatement *, CallSet *> map;
	SgStatement *wholeBodyCallSetStmt = NULL;
	int nextId = 0;
	for (int i = 0; i < recursionPointStmts.size(); i++) {
		SgStatement *stmt = recursionPointStmts[i];
		bool isBasicBlock = true;
		while (true) {
			//cout << stmt->unparseToString() << endl << endl;
			SgStatement *parent = isSgStatement(stmt->get_parent());
			ROSE_ASSERT(parent != NULL);
			if (parent == body) {
				// have first call stmt represent call set
				if (wholeBodyCallSetStmt == NULL) {
					wholeBodyCallSetStmt = stmt;
				} else {
					stmt = wholeBodyCallSetStmt;
					isBasicBlock = false;
				}
				break;
			}
			if (setContains(loopVariantControlStmts, parent)) break;
			stmt = parent;
		}
		CallSet *cset;
		if (map.count(stmt) == 0) {
			cset = new CallSet(stmt, nextId++);
			callSets.push_back(cset);
			map[stmt] = cset;
		} else {
			cset = map[stmt];
		}
		if (!isBasicBlock) cset->stmts.push_back(recursionPointStmts[i]);
		cset->addCall(recursionPoints[i]);

        cset->printCalls();
	}
}
