/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#ifndef RECURSIVECALLSET_H_
#define RECURSIVECALLSET_H_

#include "rose.h"
#include <vector>
#include <set>
#include <string>
using namespace std;

#include "Includes.h"

class CallSet {
public:
	// statements of the call set
	// this can be a single statement: basic block of calls and other statements (BH), or multiple call statements (HeatRay)
	vector<SgStatement *> stmts;

	vector<SgFunctionCallExp *> calls;
	int id;

	CallSet(SgStatement *stmt, int i) : id(i) {
		stmts.push_back(stmt);
	}

	void addCall(SgFunctionCallExp *call) {
		calls.push_back(call);
	}

	void printCalls() {
		ConstIteratorFor(i, calls, vector<SgFunctionCallExp *>) {
			cout << (*i)->unparseToString() << endl;
		}
		cout << endl;
	}
};

class RecursiveCallSet {
public:
	RecursiveCallSet(SgFunctionDeclaration *funcDec, const vector<SgInitializedName *> &variantParams) {
		analysis(funcDec->get_definition()->get_body(), variantParams, funcDec->get_name());
	}

	RecursiveCallSet(SgBasicBlock *body, const vector<SgInitializedName *> &variantParams, string orgFuncName) {
		analysis(body, variantParams, orgFuncName);
	}

	void analysis(SgBasicBlock *body, const vector<SgInitializedName *> &variantParams, string orgFuncName);

	vector<CallSet *> callSets;
};

#endif /* RECURSIVECALLSET_H_ */
