/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#include "RecursiveField.h"
#include "Globals.h"
#include "Includes.h"
#include "Utils.h"
#include "RecursiveCallSet.h"

vector<RecursiveField> RecursiveField::recursiveFields;

void RecursiveField::analyze(const RecursiveCallSet &rcs) {
	set<string> fieldNames;
	ConstIteratorFor(i, rcs.callSets, vector<CallSet *>) {
		CallSet *cs = *i;
		ConstIteratorFor(j, cs->calls, vector<SgFunctionCallExp *>) {
			SgFunctionCallExp *call = *j;
			const SgExpressionPtrList &args = call->get_args()->get_expressions();
			SgArrowExp *arrow = isSgArrowExp(args[g_nodeParamIdx]);
			ROSE_ASSERT(arrow != NULL);
			SgVarRefExp *var = isSgVarRefExp(arrow->get_rhs_operand());
			ROSE_ASSERT(var != NULL);
			fieldNames.insert(var->get_symbol()->get_name());
		}
	}
	ConstIteratorFor(i, fieldNames, set<string>) {
		recursiveFields.push_back(RecursiveField(*i));
	}

	/*
	 this way of finding fields breaks when field is not used in recursion
	 ex) parent in
	Node *parent;
	Node* left;
	Node* right;
	 */
//	Rose_STL_Container<SgNode *> fields = NodeQuery::querySubTree(
//			g_recursiveClassDec->get_definition(), V_SgVariableDeclaration);
//	for (Rose_STL_Container<SgNode*>::iterator j = fields.begin(); j != fields.end(); ++j) {
//		SgVariableDeclaration *fieldDec = isSgVariableDeclaration(*j);
//		SgInitializedNamePtrList& fieldList = fieldDec->get_variables();
//		for (SgInitializedNamePtrList::iterator k = fieldList.begin(); k != fieldList.end(); ++k) {
//			SgInitializedName *fieldVar = isSgInitializedName(*k);
//			SgPointerType *pointerType = isSgPointerType(fieldVar->get_type());
//			if (pointerType == NULL) continue;
//			SgClassType *classType = isSgClassType(pointerType->get_base_type());
//			SgClassType *recursiveType = isSgClassType(g_recursiveType->get_base_type());
//			if (classType == NULL) continue;
//			if (classType == recursiveType) {
//				cout << fieldVar->get_name() << endl;
//				recursiveFields.push_back(RecursiveField(fieldVar->get_name()));
//			}
//		}
//	}
}

int RecursiveField::getFieldId(string name) {
	//cout << name << endl;
	for (vector<RecursiveField>::iterator i = recursiveFields.begin(); i != recursiveFields.end(); i++) {
		//cout << i->name << endl;
		if (i->name == name) return i->id;
	}
	ROSE_ASSERT(false);
	return -1;
}

int RecursiveField::getFieldId(SgFunctionCallExp *call) {
	return getFieldId(getNodeVarName(call));
}
