/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#include "rose.h"
#include "Globals.h"
#include "Utils.h"
#include "MethodInfo.h"
#include "SpliceClasses.h"
#include "RecursiveCallSet.h"
#include "RecursiveField.h"

#include <iostream>
#include <fstream>
using namespace std;

using namespace SageBuilder;
using namespace SageInterface;

void SpliceClasses::generate(const RecursiveCallSet& rcs) {

	//SageInterface::insertHeader("block.h", PreprocessingInfo::after, false, g_global);
	if (g_afterClassDecsPragma == NULL) {
		SageInterface::addTextForUnparser(g_global, "#include \"splice.h\"\n", AstUnparseAttribute::e_before);
	} else {
		SageInterface::addTextForUnparser(g_afterClassDecsPragma, "#include \"splice.h\"\n", AstUnparseAttribute::e_before);
	}
	string nodeTypeName = getTypeString(g_recursiveType);
	string nodeVarName = g_nodeParam->get_name();
	string recursiveFunctionName = g_recursiveFunction->get_name();
	int numFields = RecursiveField::recursiveFields.size();
	SgFunctionCallExp *call = rcs.callSets[0]->calls[0];


	ofstream file;
	file.open ("splice.h");

	file << "class _SpliceNode {\n"
			<< "public:\n";
	vector<string> spliceNodeArgs, spliceNodeArgs2;
	for (int i = 0; i < g_methodInfo.recursiveMethodAllLoopInvariantParams.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParams[i];
		//cout << param->get_name() << "\t" << param->get_type()->class_name() << endl;
		spliceNodeArgs.push_back(getTypeString(param->get_type(), true) + " " + string(param->get_name()));
		spliceNodeArgs2.push_back(string(param->get_name()));
	}
	file << "	_SpliceNode(" << concatStrs(spliceNodeArgs) <<") {\n";
	for (int i = 0; i < spliceNodeArgs2.size(); i++) {
		file << "		this->" << spliceNodeArgs2[i] << " = " << spliceNodeArgs2[i] << ";\n";
	}
	file << "	}\n";
	file << "	~_SpliceNode() {}\n"
			<< "\n";
	file << "	void add(" << getParamsString(g_methodInfo.recursiveMethodLoopVariantParams) << ") { points.push_back(_Point(";
	vector<string> addArgs;
	for (int i = 0; i < g_methodInfo.recursiveMethodLoopVariantParams.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodLoopVariantParams[i];
		addArgs.push_back((isSgReferenceType(param->get_type()) ? "&" : "") + param->get_name());
	}
	file << concatStrs(addArgs) << ")); }\n";
	file << "	int size() { return points.size(); }\n"
			<< "	void clear() { points.clear(); }\n"
			<< "\n"
			<< "	vector<_Point> points;\n";
	for (int i = 0; i < spliceNodeArgs.size(); i++) {
		SgInitializedName *param = g_methodInfo.recursiveMethodAllLoopInvariantParams[i];
		file << "	" << spliceNodeArgs[i] << ";\n";
	}
	file << "};\n"
			<< "\n"
			<< "class _SpliceSet {\n"
			<< "public:\n"
			<< "	_SpliceSet();\n"
			<< "	~_SpliceSet();\n"
			<< "\n"
			<< "	void map_tree(" << getParamsString(g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode)
			<< ", int depth, int index);\n"
			<< "	_SpliceNode* get_node(int i) {\n"
			<< "		assert(i < num_nodes);\n"
			<< "		return splice_nodes_dest[i];\n"
			<< "	}\n"
			//<< "	" << nodeTypeName << "get_root() { return tree_map[1]; }\n"
			<< "	void do_splice();\n"
			<< "\n"
			<< "	_Block *block;\n"
			<< "	_BlockStack *stack;\n"
			<< "\n"
			<< "	static void init(int d);\n"
			<< "	static int depth_param;\n"
			<< "\n"
			<< "private:\n"
			<< "	void swap_nodes(int i);\n"
			<< "	void phase_infix(int depth1, int phase1, int depth2, int phase2);\n"
			<< "	void top_phase(int depth, int phase);\n"
			<< "\n";
	//			<< "	void bottom_upper_phase_block(int start_node_index, int depth);\n"
	//			<< "	void do_bottom_upper_phase_block(_Block *block, int target_base, _SpliceNode *sn);\n"
	if (g_spliceOptimizationLevel == Splice_NoOrder) {
		file << "	void bottom_phase_block(int start_node_index);\n"
				<< "	void do_bottom_phase_block(int target_base, _SpliceNode *sn);\n"
				<< "	void process_bottom_phase_block(" << nodeTypeName << "node, _SpliceNode *sn);\n"
				<< "	void process_top_phase_block(" << nodeTypeName << "node, int depth, _SpliceNode *sn);\n";
	} else {
		file << "	void bottom_phase_block();\n"
				<< "	void do_bottom_phase_block(int target_base, int target_index, _SpliceNode *sn);\n"
				<< "	void process_bottom_phase_block(" << nodeTypeName << "node, _SpliceNode *sn);\n"
				<< "	void process_top_phase_block(" << nodeTypeName << "node, int depth, int _index, _SpliceNode *sn);\n";
	}

	file << "\n"
			<< "	_SpliceNode **splice_nodes_src;\n"
			<< "	_SpliceNode **splice_nodes_dest;\n"
			<< "	" << nodeTypeName << "*tree_map;\n"
			<< "\n"
			//<< "	static int depth_param_half;\n"
			<< "	static int num_leaf_nodes;\n"
			<< "	static int num_nodes;\n"
			<< "};\n"
			<< "\n"
			<< "const int max_children = " << numFields << ";\n"
			<< "const int max_phases = " << numFields << ";\n"
			<< "\n"
			<< "int _SpliceSet::depth_param;\n"
			//<< "int _SpliceSet::depth_param_half;\n"
			<< "int _SpliceSet::num_leaf_nodes;\n"
			<< "int _SpliceSet::num_nodes;\n"
			<< "\n"
			<< "int int_exp(int n) {\n"
			<< "	int ret = 1;\n"
			<< "	while (n-- > 0) {\n"
			<< "		ret *= max_children;\n"
			<< "	}\n"
			<< "	return ret;\n"
			<< "}\n"
			<< "\n"
			<< "void _SpliceSet::init(int d) {\n"
			<< "	assert(d > 0);\n"
			<< "	depth_param = d;\n"
			//<< "	depth_param_half = depth_param >> 1;\n"
			<< "	num_leaf_nodes = int_exp(depth_param + 1);\n"
			<< "	num_nodes = num_leaf_nodes * max_children;\n"
			<< "	assert(num_nodes > num_leaf_nodes);	// for very large d this could overflow\n"
			<< "}\n"
			<< "\n"
			<< "_SpliceSet::_SpliceSet() {\n";
	if (g_spliceOptimizationLevel == Splice_NoOrder) {
		file << "	splice_nodes_src = new _SpliceNode*[depth_param + 2];\n"
				<< "	splice_nodes_dest = new _SpliceNode*[depth_param + 2];\n"
				<< "	for (int i = 0; i < depth_param + 2; i++) {\n"
				<< "		splice_nodes_src[i] = NULL;\n"
				<< "		splice_nodes_dest[i] = NULL;\n"
				<< "	}\n";
	} else {
		file << "	splice_nodes_src = new _SpliceNode*[num_nodes];\n"
				<< "	splice_nodes_dest = new _SpliceNode*[num_nodes];\n"
				<< "	for (int i = 0; i < num_nodes; i++) {\n"
				<< "		splice_nodes_src[i] = NULL;\n"
				<< "		splice_nodes_dest[i] = NULL;\n"
				<< "	}\n";
	}
	file << "	tree_map = new " << nodeTypeName << "[num_nodes];\n"
			<< "	block = new _Block;\n"
			<< "	stack = new _BlockStack;\n"
			<< "}\n"
			<< "\n"
			<< "_SpliceSet::~_SpliceSet() {\n";
	if (g_spliceOptimizationLevel == Splice_NoOrder) {
		file << "	for (int i = 0; i < depth_param + 2; i++) {\n";
	} else {
		file << "	for (int i = 0; i < num_nodes; i++) {\n";
	}
	file << "		if (splice_nodes_src[i] == NULL) continue;\n"
			<< "		delete splice_nodes_src[i];\n"
			<< "		delete splice_nodes_dest[i];\n"
			<< "	}\n"
			<< "	delete [] splice_nodes_src;\n"
			<< "	delete [] splice_nodes_dest;\n"
			<< "	delete [] tree_map;\n"
			<< "	delete block;\n"
			<< "	delete stack;\n"
			<< "}\n"
			<< "\n"
			<< "void _SpliceSet::map_tree(" << getParamsString(g_methodInfo.recursiveMethodAllLoopInvariantParamsIncludingNode)
			<< ", int depth, int index) {\n"
			<< "	assert(index < num_nodes);\n"
			<< "	tree_map[index] = " << nodeVarName << ";\n";
	if (g_spliceOptimizationLevel == Splice_NoOrder) {
		file << "	// for splice set stacks, can be init multiple times\n"
				<< "	if (splice_nodes_dest[depth] != NULL) return;\n"
				<< "	splice_nodes_dest[depth] = new _SpliceNode(" << concatStrs(spliceNodeArgs2) << ");\n"
				<< "	splice_nodes_src[depth] = new _SpliceNode(" << concatStrs(spliceNodeArgs2) << ");\n";
	} else {
		file << "	splice_nodes_dest[index] = new _SpliceNode(" << concatStrs(spliceNodeArgs2) << ");\n"
				<< "	splice_nodes_src[index] = new _SpliceNode(" << concatStrs(spliceNodeArgs2) << ");\n";
	}
	file << "}\n"
			<< "\n"
			<< "void _SpliceSet::swap_nodes(int i) {\n"
			//<< "	assert(i < depth_param + 2);\n"
			<< "	_SpliceNode *swap = splice_nodes_src[i];\n"
			<< "	splice_nodes_src[i] = splice_nodes_dest[i];\n"
			<< "	splice_nodes_dest[i] = swap;\n"
			<< "}\n"
			<< "\n";
	if (g_spliceOptimizationLevel == Splice_NoOrder) {
		file << "void _SpliceSet::do_splice() {\n"
				<< "	phase_infix(0, 1, 0, 1);\n"
				<< "}\n"
				<< "\n"
				<< "void _SpliceSet::phase_infix(int depth1, int phase1, int depth2, int phase2) {\n"
				<< "	if (depth2 < depth_param) {\n"
				<< "		for (int i = 0; i < max_children; i++) {\n"
				<< "			if (i == 0) {\n"
				<< "				if (tree_map[phase2 * max_children] != NULL) {\n"
				<< "					phase_infix(depth1, phase1, depth2 + 1, phase2 * max_children);\n"
				<< "				} else {\n"
				<< "					top_phase(depth1, phase1);\n"
				<< "				}\n"
				<< "			} else {\n"
				<< "				if (tree_map[phase2 * max_children + i] != NULL) {\n"
				<< "					phase_infix(depth2 + 1, phase2 * max_children + i, depth2 + 1, phase2 * max_children + i);\n"
				<< "				}\n"
				<< "			}\n"
				<< "		}\n"
				<< "	} else {\n"
				<< "		top_phase(depth1, phase1);\n"
				<< "		if (tree_map[phase2 * max_children] != NULL) {\n"
				<< "			bottom_phase_block(phase2 * max_children);\n"
				<< "		}\n"
				<< "	}\n"
				<< "}\n"
				<< "\n"
				<< "void _SpliceSet::top_phase(int depth, int node_index) {\n"
				<< "	if (depth == 0) return;\n"
				<< "\n"
				<< "	for (int i = depth; i <= depth_param + 1; i++) {\n"
				<< "		if (splice_nodes_dest[i] == NULL) continue;\n"
				<< "		swap_nodes(i);\n"
				<< "		_SpliceNode *dest = splice_nodes_dest[i];\n"
				<< "		dest->clear();\n"
				<< "	}\n"
				<< "\n"
				<< "	_SpliceNode *targetsn = splice_nodes_src[depth];\n"
				<< "	for (int i = depth; i <= depth_param + 1; i++) {\n"
				<< "		_SpliceNode *sn = splice_nodes_src[i];\n"
				<< "\n"
				<< "		for (vector<_Point>::const_iterator it = sn->points.begin() ; it < sn->points.end(); it++) {\n"
				<< "			block->add(*it);\n"
				<< "			if (block->is_full()) {\n"
				<< "				process_top_phase_block(tree_map[node_index], depth, targetsn);\n"
				<< "				block->recycle();\n"
				<< "			}\n"
				<< "		}\n"
				<< "		if (!block->is_empty()) {\n"
				<< "			process_top_phase_block(tree_map[node_index], depth, targetsn);\n"
				<< "			block->recycle();\n"
				<< "		}\n"
				<< "	}\n"
				<< "}\n"
				<< "\n"
				<< "void _SpliceSet::bottom_phase_block(int start_node_index)  {\n"
				<< "	_SpliceNode *sn = splice_nodes_dest[depth_param + 1];\n"
				<< "	if (sn == NULL) return;\n"
				<< "	int target = start_node_index;\n"
				<< "	int target_rem = target % max_children;\n"
				<< "	int target_base = target - target_rem;\n"
				<< "	assert(block->is_empty());\n"
				<< "	for (vector<_Point>::const_iterator it = sn->points.begin() ; it < sn->points.end(); it++) {\n"
				<< "		block->add(*it);\n"
				<< "		if (block->is_full()) {\n"
				<< "			do_bottom_phase_block(target_base, sn);\n"
				<< "			block->recycle();\n"
				<< "		}\n"
				<< "	}\n"
				<< "	if (!block->is_empty()) {\n"
				<< "		do_bottom_phase_block(target_base, sn);\n"
				<< "		block->recycle();\n"
				<< "	}\n"
				<< "}\n"
				<< "\n"
				<< "void _SpliceSet::do_bottom_phase_block(int target_base, _SpliceNode *sn)  {\n";
	} else {
		file << "void _SpliceSet::do_splice() {\n"
				<< "	phase_infix(0, 0, 0, 0);\n"
				<< "}\n"
				<< "\n"
				<< "void _SpliceSet::phase_infix(int depth1, int phase1, int depth2, int phase2) {\n"
				<< "	if (depth2 < depth_param) {\n"
				<< "		for (int i = 0; i < max_phases; i++) {\n"
				<< "			if (i == 0) {\n"
				<< "				phase_infix(depth1, phase1, depth2 + 1, i);\n"
				<< "			} else {\n"
				<< "				phase_infix(depth2 + 1, i, depth2 + 1, i);\n"
				<< "			}\n"
				<< "		}\n"
				<< "	} else {\n"
				<< "		top_phase(depth1, phase1);\n"
				<< "		bottom_phase_block();\n"
				<< "	}\n"
				<< "}\n"
				<< "\n"
				<< "void _SpliceSet::top_phase(int depth, int phase) {\n"
				<< "	if (depth == 0) return;\n"
				<< "\n"
				<< "	int start = int_exp(depth);\n"
				<< "	int end = start * max_children;\n"
				<< "\n"
				<< "	for (int i = start; i < num_nodes; i++) {\n"
				<< "		if (splice_nodes_dest[i] == NULL) continue;\n"
				<< "		swap_nodes(i);\n"
				<< "		_SpliceNode *dest = splice_nodes_dest[i];\n"
				<< "		dest->clear();\n"
				<< "	}\n"
				<< "	for (int i = start; i < end; i++) {\n"
				<< "		_SpliceNode *targetsn = splice_nodes_src[i];\n"
				<< "		int target = i;\n"
				<< "		int target_rem = target % max_children;\n"
				<< "		int target_base = target - target_rem;\n"
				<< "		int next_node;\n";

		//				<< "		if (target_rem == 0) {\n"
		//				<< "			next_node = target_base + 1;\n"
		//				<< "		} else {\n"
		//				<< "			next_node = target_base + 0;\n"
		//				<< "		}\n"
		vector<IfPart> ifparts;
		for (int i = 1; i < numFields; i++) {
			IfPart ifpart;
			stringstream sstm;
			sstm << "phase == " << i;
			ifpart.cond = sstm.str();
			vector<IfPart> innerIfparts;
			for (int j = 0; j < rcs.callSets.size(); j++) {
				CallSet *cs = rcs.callSets[j];
//				cout << i << " " << j << endl;
//				cs->printCalls();
				int prevId = RecursiveField::getFieldId(cs->calls[i - 1]);
				int curId = RecursiveField::getFieldId(cs->calls[i]);
				IfPart innerIfpart;
				stringstream sstm2, sstm3;
				sstm2 << "target_rem == " << prevId;
				innerIfpart.cond = sstm2.str();
				sstm3 << "next_node = target_base + " << curId << ";\n";
				innerIfpart.lines.push_back(sstm3.str());
				innerIfparts.push_back(innerIfpart);
			}
			ifpart.lines.push_back(getIfStr(innerIfparts, "\t"));
			ifparts.push_back(ifpart);
		}
		file << getIfStr(ifparts, "\t\t");
		file << "		int sub_start = i;\n"
				<< "		int sub_end = i + 1;\n"
				<< "		while (sub_start < num_nodes) {\n"
				<< "			for (int k = sub_start; k < sub_end; k++) {\n"
				<< "				_SpliceNode * sn = splice_nodes_src[k];\n"
				<< "				if (sn == NULL) continue;\n"
				<< "				for (vector<_Point>::iterator it = sn->points.begin() ; it < sn->points.end(); it++) {\n"
				<< "					block->add(*it);\n"
				<< "					if (block->is_full()) {\n"
				<< "						process_top_phase_block(tree_map[next_node], depth, next_node, targetsn);\n"
				<< "						block->recycle();\n"
				<< "					}\n"
				<< "				}\n"
				<< "			}\n"
				<< "			sub_start *= max_children;\n"
				<< "			sub_end *= max_children;\n"
				<< "		}\n"
				<< "		if (!block->is_empty()) {\n"
				<< "			process_top_phase_block(tree_map[next_node], depth, next_node, targetsn);\n"
				<< "			block->recycle();\n"
				<< "		}\n"
				<< "	}\n"
				<< "}\n"
				<< "\n"
				<< "void _SpliceSet::bottom_phase_block()  {\n"
				<< "	for (int i = num_leaf_nodes; i < num_nodes; i++) {\n"
				<< "		_SpliceNode * sn = splice_nodes_dest[i];\n"
				<< "		if (sn == NULL) continue;\n"
				<< "		int target = i;\n"
				<< "		int target_rem = target % max_children;\n"
				<< "		int target_base = target - target_rem;\n"
				<< "		for (vector<_Point>::iterator it = sn->points.begin() ; it < sn->points.end(); it++) {\n"
				<< "			block->add(*it);\n"
				<< "			if (block->is_full()) {\n"
				<< "				do_bottom_phase_block(target_base, target_rem, sn);\n"
				<< "				block->recycle();\n"
				<< "			}\n"
				<< "		}\n"
				<< "		if (!block->is_empty()) {\n"
				<< "			do_bottom_phase_block(target_base, target_rem, sn);\n"
				<< "			block->recycle();\n"
				<< "		}\n"
				<< "	}\n"
				<< "}\n"
				<< "\n"
				<< "void _SpliceSet::do_bottom_phase_block(int target_base, int target_rem, _SpliceNode *sn)  {\n";
	}
	vector<IfPart> ifparts;
	for (int i = 0; i < rcs.callSets.size(); i++) {
		CallSet *cs = rcs.callSets[i];
		IfPart ifpart;
		stringstream sstm;
		sstm << "target_rem == " << cs->id;
		ifpart.cond = sstm.str();
		for (int j = 0; j < cs->calls.size(); j++) {
			string nodeVarName = getNodeVarName(cs->calls[j]);
			int fieldId = RecursiveField::getFieldId(nodeVarName);
			stringstream linesstm;
			linesstm << "	process_bottom_phase_block(tree_map[target_base + " << fieldId << "], sn);\n";
			ifpart.lines.push_back(linesstm.str());
		}
		ifparts.push_back(ifpart);
	}
	file << getIfStr(ifparts, "\t");
	file << "}\n"
			<< "\n";

	/* forward declare block and blockSplice recursive functions
	void compute_force_recursive_block(class Node *node,float dsq,float epssq,class _BlockStack *_stack,int _depth);
	void compute_force_recursive_blockSplice(class Node *node,float dsq,float epssq,class _BlockStack *_stack,int _depth,class _SpliceSet *_spliceSet);
	 */

	vector<string> params, spliceParams;
	const SgInitializedNamePtrList &funcParams = g_recursiveFunction->get_args();
	for (int i = 0; i < funcParams.size(); i++) {
		if (vecContains(g_methodInfo.recursiveMethodAllLoopVariantParamsIdxs, i)) continue;
		SgInitializedName *param = funcParams[i];
		params.push_back(getTypeString(param->get_type()) + " " + param->get_name());
		spliceParams.push_back(getTypeString(param->get_type()) + " " + param->get_name());
	}
	params.push_back("_BlockStack *_stack");
	params.push_back("int _depth");
	spliceParams.push_back("_BlockStack *_stack");
	spliceParams.push_back("int _depth");
	if (g_spliceOptimizationLevel != Splice_NoOrder) {
		spliceParams.push_back("int _index");
	}
	spliceParams.push_back("_SpliceSet *_spliceSet");

	file << "void " << recursiveFunctionName << "_block(" << concatStrs(params) << ");\n";
	file << "void " << recursiveFunctionName << "_blockSplice(" << concatStrs(spliceParams) << ");\n";
	file << "\n";

	// process_bottom_phase_block, process_top_phase_block

	vector<string> args, spliceArgs;
	const SgExpressionPtrList &callArgs = call->get_args()->get_expressions();
	//cout << call->unparseToString() << endl;
	for (int i = 0; i < callArgs.size(); i++) {
		if (vecContains(g_methodInfo.recursiveMethodAllLoopVariantParamsIdxs, i)) continue;
		if (i == g_nodeParamIdx) {
			args.push_back("node");
			spliceArgs.push_back("node");
		}
		else {
			SgVarRefExp *varRef = isSgVarRefExp(callArgs[i]);
			//ROSE_ASSERT(varRef != NULL);
			if (varRef == NULL) continue;
			args.push_back("sn->" + varRef->get_symbol()->get_name());
			spliceArgs.push_back("sn->" + varRef->get_symbol()->get_name());
		}
	}
	args.push_back("stack");
	args.push_back("0");
	spliceArgs.push_back("stack");
	spliceArgs.push_back("depth");
	if (g_spliceOptimizationLevel != Splice_NoOrder) {
		spliceArgs.push_back("index");
	}
	spliceArgs.push_back("this");

	file << "void _SpliceSet::process_bottom_phase_block(" << nodeTypeName << "node, _SpliceNode *sn)  {\n"
//			<< "	if (node == NULL) return;\n"
			<< "	stack->get(0)->block = block;\n";
	file << "	";
	if (g_recursiveMemberFunction != NULL) file << "_Block::root_instance->";
	file << recursiveFunctionName << "_block(" << concatStrs(args) << ");\n";
	if (g_useSOA) {
	file << "	block->copy_back();\n";
	}
	file << "}\n"
			<< "\n";
	if (g_spliceOptimizationLevel == Splice_NoOrder) {
		file << "void _SpliceSet::process_top_phase_block(" << nodeTypeName << "node, int depth, _SpliceNode *sn)  {\n";
	} else {
		file << "void _SpliceSet::process_top_phase_block(" << nodeTypeName << "node, int depth, int index, _SpliceNode *sn)  {\n";
	}
//	file << "	if (node == NULL) return;\n"
			file << "	stack->get(depth)->block = block;\n";
	file << "	";
	if (g_recursiveMemberFunction != NULL) file << "_Block::root_instance->";
	file << recursiveFunctionName << "_blockSplice(" << concatStrs(spliceArgs) << ");\n";
	if (g_useSOA) {
	file << "	block->copy_back();\n";
	}
	file << "}\n";

	file.close();
}
