/*
SIMTree, a framework to automatically vectorize tree traversal algorithms.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2013, School of Electrical and Computer Engineering, Purdue University.
 */

#ifndef UTILS_H_
#define UTILS_H_

#include "rose.h"
#include <string>
#include <vector>
#include <set>
using namespace std;

struct IfPart {
	string cond;
	vector<string> lines;
};

string getIfStr(const vector<IfPart>& ifparts, string prefix = "");

string getNodeVarName(SgFunctionCallExp *call);

SgType *getBaseType(SgType *t);

SgFunctionCallExp *getEnclosingFunctionCall(SgNode *node);


string concatStrs(const vector<string> &strs);

template <class T>
inline bool vecContains(const vector<T> &v, T i) {
	return (find(v.begin(), v.end(), i) != v.end());
}

template <class T>
inline bool setContains(const set<T> &v, T i) {
	return (find(v.begin(), v.end(), i) != v.end());
}

string getTypeString(const SgType* type, bool refToPointer = false);

string getArgsString(const vector<SgVarRefExp *> &args);
string getArgsString(const vector<SgInitializedName *> &args);
string getParamsString(const vector<SgInitializedName *> &args, bool refToPointer = false);
string getArgsStringWithPrefix(const vector<SgInitializedName *> &args, string prefix);

SgExprListExp *getArgsExprList(const vector<SgVarRefExp *> &v);
SgExprListExp *getArgsExprList(const vector<SgInitializedName *> &v, SgScopeStatement *scope);

bool hasVarUse(SgStatement *stmt, string varName);

string getClassForwardDeclareString();

void addForwardFuncDecIfNecessary(SgFunctionDeclaration *orgFunc, SgFunctionDeclaration *newFunc);

bool exprContainsVarName(SgExpression *expr, const set<string> &names);
bool exprContainsVarName(SgExpression *expr, const vector<string> &names);
bool gatherLoopVariantStmts(set<SgStatement *> *loopVariantControlStmts, set<string> *loopVariantArgsPropagation,
		SgStatement *stmt, const vector<SgStatement *> &recursionPointStmts);
void removeLoopVariantStmts(vector<SgStatement *> *toRemoveStmts, set<string> *loopVariantArgsPropagation,
		SgStatement *stmt, const vector<SgStatement *> &recursionPointStmts);
void getRecursionPoints(SgFunctionDeclaration *funcDec, vector<SgFunctionCallExp *> *recursionPoints);
void getRecursionPoints(SgBasicBlock *body, vector<SgFunctionCallExp *> *recursionPoints, string orgFuncName);

SgAssignInitializer *getVarDecInit(SgVariableDeclaration *varDec, SgInitializedName **var = NULL);
SgInitializedName *getVarDecInitName(SgVariableDeclaration *varDec);

SgFunctionDefinition *getFunctionDefintion(string name);
SgClassDefinition *getClassDefintion(string name);


// param must be a pointer
// returns SgNamedType if param is pointer of named type, NULL if not named type e.g., int
SgNamedType *getNamedType(SgInitializedName *param);

#endif /* UTILS_H_ */
